"""Piecewise coupling cells
Coupling cells whose transform f is defined as the primitive of a function g
such that
* f(0) = 0
* f(1) = 1
* g(y,t^N) > 0
* g is a piecewise-simple function

e.g.: g can be a positive, piecewise constant function, whose step heights are
the outputs of a neural network taking the y^N as input, properly normalized to
integrate to 1.

Throughout this sub-package, "PW" stands for piecewise.
"""


from better_abc import abstract_attribute
from ..general_flow import GeneralFlow

class AnalyticFlow(GeneralFlow):
    def __init__(self,d):
        super(AnalyticFlow, self).__init__(d=d)
        self.flow_ = abstract_attribute()

    def flow(self, x):
        return self.flow_(x)


class InvertibleAnalyticFlow(GeneralFlow):
    def __init__(self,d):
        super(InvertibleAnalyticFlow, self).__init__(d=d)

        # I'd like to have these but because of how PyTorch registers modules
        # This does not work
        #self.flow_ = abstract_attribute()
        #self.inverse_flow_ = abstract_attribute()

        self.inverse = False

    def runs_forward(self):
        """Check the running mode: True if forward and False if backward/inverse"""
        return not self.inverse

    def invert(self):
        self.inverse = not self.inverse

    def flow(self, x):
        if self.inverse:
            return self.inverse_flow_.flow(x)

        return self.flow_.flow(x)

    def transform_and_compute_jacobian(self, xj):
        if self.inverse:
            return self.backward_flow.transform_and_compute_jacobian(xj)

        return self.forward_flow.transform_and_compute_jacobian(xj)



from .analytic_flow import AnalyticFlow,InvertibleAnalyticFlow
import torch


class AnalyticSigmoid(AnalyticFlow):
    def __init__(self,d):
        super(AnalyticSigmoid, self).__init__(d=d)
        self.flow_ = torch.nn.Sigmoid()

    def transform_and_compute_jacobian(self, xj):
        x = xj[..., :-1]
        logj = xj[..., -1]
        yj = torch.zeros_like(xj).to(xj.device)
        yj[..., :-1] = self.flow(x)

        yj[..., -1] = logj+torch.sum(torch.log(yj[..., :-1] * (1 - yj[..., :-1])), dim=-1)

        return yj


def inverse_sigmoid(x):
    return torch.log(x / (1 - x))


class AnalyticInverseSigmoid(AnalyticFlow):
    def __init__(self,d):
        super(AnalyticInverseSigmoid, self).__init__(d=d)
        self.flow_ = inverse_sigmoid

    def transform_and_compute_jacobian(self, xj):
        x = xj[..., :-1]
        logj = xj[..., -1]
        yj = torch.zeros_like(xj).to(xj.device)
        yj[..., :-1] = self.flow(x)

        yj[..., -1] = logj-torch.sum(torch.log(x * (1 - x)), dim=-1)

        return yj


class InvertibleAnalyticSigmoid(InvertibleAnalyticFlow):
    def __init__(self,d):
        super(InvertibleAnalyticSigmoid, self).__init__(d=d)
        self.forward_flow = AnalyticSigmoid(d)
        self.backward_flow = AnalyticInverseSigmoid(d)



import torch

from zunis.models.flows.general_flow import GeneralFlow


class InvertibleSequentialFlow(GeneralFlow):
    def __init__(self,d,flows):
        super(InvertibleSequentialFlow, self).__init__(d=d)

        self.flows = torch.nn.ModuleList(flows)

        for flow in flows:
            try:
                if not flow.runs_forward():
                    flow.invert()
            except AttributeError as e:
                print("All flows in an InvertibleSequentialFlow must be invertible")
                raise
        self.inverse = False

    def invert(self):
        self.inverse = not self.inverse
        for flow in self.flows:
            flow.invert()

    def flow(self, x):
        output = x
        if self.inverse:
            for f in self.flows[::-1]:
                output = f.flow(output)
        else:
            for f in self.flows:
                output = f.flow(output)

        return output


    def transform_and_compute_jacobian(self, xj):
        output = xj
        if self.inverse:
            for f in self.flows[::-1]:
                output = f.transform_and_compute_jacobian(output)
        else:
            for f in self.flows:
                output = f.transform_and_compute_jacobian(output)

        return output


"""High level interface for integration"""
import torch
import logging

from .flat_survey_integrator import FlatSurveySamplingIntegrator
from .dkltrainer_integrator import DKLAdaptiveSurveyIntegrator
from .adaptive_survey_integrator import ForwardSurveySamplingIntegrator, VarianceAdaptiveSurveyIntegrator
from .fixed_sample_integrator import FixedSampleSurveyIntegrator

from zunis.training.weighted_dataset.stateful_trainer import StatefulTrainer

logger = logging.getLogger(__name__)


def Integrator(f, d, survey_strategy="forward_flat_init", n_iter=10, n_iter_survey=None, n_iter_refine=None,
               n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
               device=torch.device("cpu"), verbosity=None, trainer_verbosity=None,
               loss="dkl", flow="pwquad", trainer=None, trainer_options=None, flow_options=None):
    """High level integration API

    This is a factory method that instantiates the relevant Integrator subclass based on the options

    Parameters
    ----------
    f: function
        the function to integrate
    d: int
        dimensionality of the integration space
    survey_strategy: str
        how points are sampled during the survey step: one of `'flat'`, `'forward'`, `'adaptive_dkl'`, `'adaptive_variance'`,
        `'forward_flat_init'`, `'fixed_sample'`
    n_iter: int
        general number of iterations - ignored for survey/refine if n_iter_survey/n_inter_refine is set
    n_iter_survey: int
        number of iterations for the survey stage
    n_iter_refine: int
        number of iterations for the refine stage
    n_points:
        general number of points per iteration - ignored for survey/refine if n_points_survey/n_points_refine is set
    n_points_survey: int
        number of points per iteration for the survey stage
    n_points_refine: int
        number of points per iteration for the refine stage
    use_survey: bool
        whether to use the points generated during the survey to compute the final integral
        not recommended due to uncontrolled correlations in error estimates
    device:
        pytorch device on which to run
    verbosity: int
        verbosity level of the integrator
    trainer_verbosity:
        verbosity level of the trainer
    loss: str or function
        loss function used by the trainer
    flow: str or flow model
        normalizing flow model to use for importance sampling
    trainer: None or :py:class:`zunis.training.weighted_dataset.weighted_dataset_trainer.BasicStatefulTrainer`
        optional argument to provide a full trainer object - overrides any other trainer or model setup argument
    trainer_options: dict
        dictionary of options to pass to the :py:class:`zunis.training.weighted_dataset.stateful_trainer.StatefulTrainer`
        created if `trainer` is None
    flow_options: dict
        dictionary of options to pass to the :py:class:`zunis.models.flows.sequential.repeated_cell.RepeatedCellFlow`
        created by the stateful trainer if `trainer` is None

    Returns
    -------
       subclass of :py:class:`zunis.models.flows.sequential.repeated_cell.RepeatedCellFlow`
    """
    if trainer is None:
        if trainer_options is None:
            trainer_options = dict()
        trainer = StatefulTrainer(d=d, loss=loss, flow=flow, device=device, flow_options=flow_options,
                                  **trainer_options)

    # TODO work out the issue for the variance training with RealNVP
    if loss == "variance" and flow == "realnvp":
        logger.warning(f"======================================================")
        logger.warning(f"Using loss {loss} and flow {flow}. This usually fails.")
        logger.warning(f"======================================================")

    if survey_strategy == "flat":
        return FlatSurveySamplingIntegrator(f=f, trainer=trainer, d=d, n_iter=n_iter, n_iter_survey=n_iter_survey,
                                            n_iter_refine=n_iter_refine,
                                            n_points=n_points, n_points_survey=n_points_survey,
                                            n_points_refine=n_points_refine, use_survey=use_survey,
                                            device=device, verbosity=verbosity,
                                            trainer_verbosity=trainer_verbosity)

    if survey_strategy == "adaptive_dkl":
        if trainer is None:
            assert loss == "dkl", "The adaptive DKL strategy must be used with the DKL loss"
        return DKLAdaptiveSurveyIntegrator(f=f, trainer=trainer, d=d, n_iter=n_iter, n_iter_survey=n_iter_survey,
                                           n_iter_refine=n_iter_refine,
                                           n_points=n_points, n_points_survey=n_points_survey,
                                           n_points_refine=n_points_refine, use_survey=use_survey,
                                           device=device, verbosity=verbosity,
                                           trainer_verbosity=trainer_verbosity)

    if survey_strategy == "adaptive_variance":
        if trainer is None:
            assert loss == "variance", "The adaptive variance strategy must be used with the variance loss"
        return VarianceAdaptiveSurveyIntegrator(f=f, trainer=trainer, d=d, n_iter=n_iter, n_iter_survey=n_iter_survey,
                                                n_iter_refine=n_iter_refine,
                                                n_points=n_points, n_points_survey=n_points_survey,
                                                n_points_refine=n_points_refine, use_survey=use_survey,
                                                device=device, verbosity=verbosity,
                                                trainer_verbosity=trainer_verbosity)

    if survey_strategy == 'forward':
        return ForwardSurveySamplingIntegrator(f=f, trainer=trainer, d=d, n_iter=n_iter, n_iter_survey=n_iter_survey,
                                               n_iter_refine=n_iter_refine,
                                               n_points=n_points, n_points_survey=n_points_survey,
                                               n_points_refine=n_points_refine, use_survey=use_survey,
                                               device=device, verbosity=verbosity,
                                               trainer_verbosity=trainer_verbosity,
                                               sample_flat_once=False)

    if survey_strategy == 'forward_flat_init':
        return ForwardSurveySamplingIntegrator(f=f, trainer=trainer, d=d, n_iter=n_iter, n_iter_survey=n_iter_survey,
                                               n_iter_refine=n_iter_refine,
                                               n_points=n_points, n_points_survey=n_points_survey,
                                               n_points_refine=n_points_refine, use_survey=use_survey,
                                               device=device, verbosity=verbosity,
                                               trainer_verbosity=trainer_verbosity,
                                               sample_flat_once=True)

    if survey_strategy == 'fixed_sample':
        return FixedSampleSurveyIntegrator(f=f, trainer=trainer, d=d, n_iter=n_iter, n_iter_survey=n_iter_survey,
                                           n_iter_refine=n_iter_refine,
                                           n_points=n_points, n_points_survey=n_points_survey,
                                           n_points_refine=n_points_refine, use_survey=use_survey,
                                           device=device, verbosity=verbosity,
                                           trainer_verbosity=trainer_verbosity)

    raise ValueError("""No valid survey strategy was provided. Allowed strategies are:
     ['flat', 'adaptive_dkl', 'adaptive_variance', 'forward', 'forward_flat_init', 'fixed_sample']""")



"""Model utilities"""

import torch


class Reshift(torch.nn.Module):
    """Un-trainable activation: shift and scale data

    The input x is transformed as x*scale + offset. Scale and offset are untrainable scalars.
    """
    def __init__(self, scale=2., offset=-1.):
        """

        Parameters
        ----------
        scale: float
        offset: float
        """
        super(Reshift, self).__init__()
        self.scale = torch.nn.Parameter(data=torch.scalar_tensor(scale), requires_grad=False)
        self.offset = torch.nn.Parameter(data=torch.scalar_tensor(offset), requires_grad=False)

    def forward(self, x):
        return x*self.scale + self.offset



"""Implementation of basic integrator functions"""
import numpy as np
import pandas as pd
import torch
from zunis.integration.integratorAPI import SurveyRefineIntegratorAPI
from zunis.training.weighted_dataset.weighted_dataset_trainer import BasicTrainer


class BaseIntegrator(SurveyRefineIntegratorAPI):
    """Base abstract class that implements common functionality"""

    @staticmethod
    def empty_history():
        """Create an empty history object"""
        return pd.DataFrame({
            "integral": pd.Series([], dtype="float"),
            "error": pd.Series([], dtype="float"),
            "n_points": pd.Series([], dtype="int"),
            "phase": pd.Series([], dtype="str")
        })

    def __init__(self, f, trainer, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 verbosity=None, trainer_verbosity=None, **kwargs):
        """

        Parameters
        ----------
        f: function
            the function to integrate
        n_iter: int
            general number of iterations - ignored for survey/refine if n_iter_survey/n_inter_refine is set
        n_iter_survey: int
            number of iterations for the survey stage
        n_iter_refine: int
            number of iterations for the refine stage
        n_points:
            general number of points per iteration - ignored for survey/refine if n_points_survey/n_points_refine is set
        n_points_survey: int
            number of points per iteration for the survey stage
        n_points_refine: int
            number of points per iteration for the refine stage
        use_survey: bool
            whether to use the points generated during the survey to compute the final integral
            not recommended due to uncontrolled correlations in error estimates
        verbosity: int
            verbosity level of the integrator
        """
        super(BaseIntegrator, self).__init__(verbosity=verbosity, **kwargs)
        self.f = f

        self.n_iter_survey = n_iter_survey if n_iter_survey is not None else n_iter
        self.n_iter_refine = n_iter_refine if n_iter_refine is not None else n_iter
        self.n_points_survey = n_points_survey if n_points_survey is not None else n_points
        self.n_points_refine = n_points_refine if n_points_refine is not None else n_points

        self.use_survey = use_survey


        assert isinstance(trainer, BasicTrainer), "This integrator relies on the BasicTrainer API"
        self.model_trainer = trainer
        self.model_trainer.set_verbosity(trainer_verbosity)

        self.integration_history = self.empty_history()

    def initialize(self, **kwargs):
        self.integration_history = self.empty_history()

    def initialize_survey(self, **kwargs):
        pass

    def initialize_refine(self, **kwargs):
        pass

    def sample_refine(self, *, n_points=None, f=None, **kwargs):
        if n_points is None:
            n_points = self.n_points_refine
        if f is None:
            f = self.f

        xj = self.model_trainer.sample_forward(n_points)
        x = xj[:, :-1]
        px = torch.exp(-xj[:, -1])
        fx = f(x)

        return x, px, fx

    def process_survey_step(self, sample, integral, integral_var, training_record, **kwargs):
        x, px, fx = sample
        n_points = x.shape[0]
        self.integration_history = self.integration_history.append(
            {"integral": integral,
             "error": (integral_var / n_points) ** 0.5,
             "n_points": n_points,
             "phase": "survey",
             "training record": training_record},
            ignore_index=True
        )
        self.logger.info(f"Integral: {integral:.3e} +/- {(integral_var / n_points) ** 0.5:.3e}")

    def process_refine_step(self, sample, integral, integral_var, **kwargs):
        x, px, fx = sample
        n_points = x.shape[0]
        self.integration_history = self.integration_history.append(
            {"integral": integral,
             "error": (integral_var / n_points) ** 0.5,
             "n_points": n_points,
             "phase": "refine"},
            ignore_index=True
        )

        self.logger.info(f"Integral: {integral:.3e} +/- {(integral_var / n_points) ** 0.5:.3e}")

    def finalize_survey(self, **kwargs):
        pass

    def finalize_refine(self, **kwargs):
        pass

    def finalize_integration(self, use_survey=None, **kwargs):
        if use_survey is None:
            use_survey = self.use_survey

        if use_survey:
            data = self.integration_history
        else:
            data = self.integration_history.loc[self.integration_history["phase"] == "refine"]

        result = (data["integral"] * data["n_points"]).sum()
        result /= data["n_points"].sum()

        error = np.sqrt(((data["error"] * data["n_points"]) ** 2).sum() / (data["n_points"].sum()) ** 2)

        self.logger.info(f"Final result: {float(result):.5e} +/- {float(error):.5e}")

        return float(result), float(error), self.integration_history

    def survey(self, n_survey_steps=None, **kwargs):
        if n_survey_steps is None:
            n_survey_steps = self.n_iter_survey
        super(BaseIntegrator, self).survey(n_survey_steps=n_survey_steps, **kwargs)

    def refine(self, n_refine_steps=None, **kwargs):
        if n_refine_steps is None:
            n_refine_steps = self.n_iter_refine
        super(BaseIntegrator, self).refine(n_refine_steps=n_refine_steps, **kwargs)

    def integrate(self, n_survey_steps=None, n_refine_steps=None, **kwargs):
        """Perform the integration"""
        return super(BaseIntegrator, self).integrate(n_survey_steps=n_survey_steps,
                                                                 n_refine_steps=n_refine_steps, **kwargs)



"""This module implements an abstract class for flows defined as the application of a
differentiable pytorch function on an input point. In train mode, the last entry of the
input is the logarithm of the inverse PDF at that point.
"""
import torch
from better_abc import abstract_attribute
from zunis.models.flows.general_flow import GeneralFlow


class GeneralBackpropJacobianFlow(GeneralFlow):
    """Abstract class for a flow defined as a general differentiable Pytorch function
    This implements the universal transform_and_compute_jacobian behavior:
    y = flow(x), - log(j(y)) = - log(j(x)) + log(det(dy_i/dx_j))
    The jacobian is computed naively by using autograd.
    """
    def __init__(self, *, d):
        super(GeneralBackpropJacobianFlow, self).__init__(d=d)
        # A backprop Jacobian flow must implement an attribute flow_
        # which computes the transformation in a differentiable way
        self.flow_ = abstract_attribute()

    def flow(self, x):
        """For a backprop_flow, the transformation is just a pytorch module"""
        return self.flow_(x)

    def transform_and_compute_jacobian(self, xj):
        """Compute the flow transformation and its Jacobian using pytorch.autograd"""
        x = xj[:, :self.d].detach()
        log_j = xj[:, -1]

        x.requires_grad = True
        y = self.flow_(x)

        n_batch = xj.shape[0]

        jx = torch.zeros(n_batch, self.d, self.d).to(log_j.device)
        directions = torch.eye(self.d).to(log_j).unsqueeze(0).repeat(n_batch, 1, 1)

        for i in range(self.d):
            jx[:, i, :] = torch.autograd.grad(y, x, directions[:, i, :],
                                              allow_unused=True, create_graph=True, retain_graph=True)[0]
        x.requires_grad = False
        x.grad = None

        log_det_j = torch.log(torch.abs(torch.det(jx)))
        return torch.cat([y.detach(), (log_j + log_det_j).unsqueeze(1)], 1)



"""Simple flows with an easily computable jacobian - useful for tests and checks"""
import torch
from .general_backprop_j_flow import GeneralBackpropJacobianFlow


class LinearFlow(GeneralBackpropJacobianFlow):
    """Flow transformation as a pure matrix multiplication
    Expected jacobian: det(self.flow.weight)"""
    def __init__(self, *, d):
        super(LinearFlow, self).__init__(d=d)

        self.flow_ = torch.nn.Linear(d, d, bias=False)

    def weight_init_identity_(self,std=None):
        """Initialize weights as eye + normal(0,std)
        Essentially realizing a resnet-like layer
        """
        if std is None:
            std_ = 0.1/self.d
        else:
            std_ = std
        torch.nn.init.normal_(self.flow_.weight, std=std_)
        self.flow_.weight.data += torch.eye(self.d).to(self.flow_.weight.data.device)


class SigmoidFlow(GeneralBackpropJacobianFlow):
    """Flow transformation as an element-wise sigmoid application
    Expected jacobian: sigmoid(x)*(1-sigmoid(x))"""
    def __init__(self, *, d):
        super(SigmoidFlow, self).__init__(d=d)

        layers = [torch.nn.Sigmoid()]
        self.flow_ = (torch.nn.Sequential(*layers))



from dictwrapper import DictWrapper


class TrainingRecord(DictWrapper):
    """Dictionary-like object to hold records about a training run"""

    def __init__(self, metrics=None, metadata=None, config=None, alpha=.3, checkpoint=None, **kwargs):
        super(TrainingRecord, self).__init__()

        self.update({
            "metrics": {
                "loss": []
            },
            "epochs": [],
            "step": 0,
            "loss": None,
            "best_loss": None
        })

        # Smoothing parameter for the loss running average
        self.alpha = alpha

        if metrics is not None:
            for metric in metrics:
                self["metrics"][metric] = []

        if metadata is not None:
            self["metadata"] = metadata

        if config is not None:
            self["config"] = config

        if checkpoint is not None:
            # check if the checkpoint is a valid filepath
            # delete it if it is not empty
            open(str(checkpoint), "w+").close()
            self["checkpoint"] = str(checkpoint)

        self.update(kwargs)

    @property
    def metrics(self):
        return self["metrics"]

    @property
    def losses(self):
        return self["metrics"]["loss"]

    @property
    def loss(self):
        return self["loss"]

    @property
    def step(self):
        return self["step"]

    @property
    def epoch(self):
        return len(self["epochs"])

    def log_metric(self, value, metric):
        self["metrics"][metric].append(value)

    def log_loss(self, loss):
        self.log_metric(loss, "loss")
        if self["loss"] is None:
            self["loss"] = loss
            self["best_loss"] = loss
        else:
            self["loss"] = self["loss"] * (1 - self.alpha) + self.alpha * loss
            self["best_loss"] = min(self["best_loss"], self["loss"])

    def next_step(self):
        self["step"] += 1

    def new_epoch(self):
        self["epochs"].append(self.step)

    def __repr__(self):
        step = self["step"]
        epoch = len(self["epochs"])
        metrics = [k for k in self["metrics"] if k != "loss"]
        if len(metrics) > 0:
            metric_report = ", metrics: " + str(metrics)
        else:
            metric_report = ""
        try:
            loss_report = ", loss: " + str(self.loss)
        except IndexError:
            loss_report = ""

        return f"{self.__class__.__name__}(step: {step}, epoch: {epoch}{loss_report}{metric_report})"

    def __str__(self):
        return self.__repr__()



"""Optimization of invertible flows in the weighted dataset problem using the DKL loss

Reminder: we have a dataset (x,p(x),f(x)) such that
- x ~ p(x)
- we want to learn a model that draws points according to f(x),
which is positive, and known up to normalization

We want to optimize a function q(x) such that doing importance sampling to compute f(x)
with it minimizes the variance.

The variance of the importance sampling estimator is our proto-loss

pL(f,q) =  ∫ dx q(x) (f(x)/q(x))^2 - (∫ dx q(x) f(x)/q(x))^2
       =  ∫ dx (f(x)^2/q(x)) - I(f)^2

where I(f) is the integral we want to compute and is independent of q, so our real loss is

L(f,q) = ∫ dx f(x)^2/q(x)

Which we further can compute using importance sampling from p(x):

L(f,q) = ∫ dx p(x) f(x)^2/q(x)/p(x)

Which we can compute from our dataset as the expectation value

L(f,q) = E(f(x)^2/(q(x) p(x)), x~p(x)
"""

import logging
import torch
from .weighted_dataset_trainer import BasicTrainer, BasicStatefulTrainer

logger = logging.getLogger(__name__)


def weighted_variance_loss(fx, px, logqx):
    """Proxy variance loss for the integral of a function f using importance sampling from q,
    but where the variance is estimated with importance sampling from p.


    We want to optimize a function q(x) such that doing importance sampling to compute f(x)
    with it minimizes the variance.

    The variance of the importance sampling estimator is our proto-loss

    pL(f,q) =  ∫ dx q(x) (f(x)/q(x))^2 - (∫ dx q(x) f(x)/q(x))^2
           =  ∫ dx (f(x)^2/q(x)) - I(f)^2

    where I(f) is the integral we want to compute and is independent of q, so our real loss is

    L(f,q) = ∫ dx f(x)^2/q(x)

    Which we further can compute using importance sampling from p(x):

    L(f,q) = ∫ dx p(x) f(x)^2/q(x)/p(x)

    Which we can compute from our dataset as the expectation value
    """
    return torch.mean(fx ** 2 / (px * torch.exp(logqx)))


class BasicVarTrainer(BasicTrainer):
    """Basic trainer based on the variance loss"""
    def __init__(self, flow, latent_prior):
        super(BasicVarTrainer, self).__init__(flow, latent_prior)
        self.loss = weighted_variance_loss


class BasicStatefulVarTrainer(BasicStatefulTrainer):
    """Basic stateful trainer based on the variance loss"""
    def __init__(self, flow, latent_prior, **kwargs):
        super(BasicStatefulVarTrainer, self).__init__(flow, latent_prior, **kwargs)
        self.loss = weighted_variance_loss



"""Tools to create configuration files for ZuNIS objects"""
import ruamel.yaml as yaml

from .configuration import Configuration
from .loaders import get_default_integrator_config


def create_integrator_config_file(filepath="integator_config.yaml", base_config=None, force=False, **kwargs):
    """Create a config file for an integrator.

    Parameters
    ----------
    filepath: str
        path of the file to which a config file will be saved
    base_config: str or None
        path of the source config file on which to base the new config.
        If none, use the default provided with the library
    force: bool, False
        whether to overwrite an existing target file
    **kwargs:
        options that will be set in the new target file. If the option does not already exist in the base config file,
        the option is created at the top level. Otherwise, the existing option is updated even if it is nested
        (see :py:class:`zunis.utils.config.configuration.Configuration`).

    """
    if base_config is None:
        config = get_default_integrator_config()
    else:
        config = Configuration.from_yaml(base_config)

    for arg in kwargs:
        config[arg] = kwargs[arg]

    config = config.as_dict()

    mode = "w" if force else "x"

    yaml.dump(config, open(filepath, mode=mode))



"""Neural network as a flow"""
import torch
from .general_backprop_j_flow import GeneralBackpropJacobianFlow
from zunis.models.layers.activations import NormBiTanh


class NNFlow(GeneralBackpropJacobianFlow):
    """Flow defined as a neural network mapping R^d to R^d.
    Bijectivity is not guaranteed but likely: each linear layer *can* have rank >= d
    and each activation layer is bijective.

    With weights taken randomly from any reasonable definition (flat Glorot, normal Glorot),
    the probability that the matrices are less than maximal rank is 0 so the transformation is
    nearly always bijective.

    The layers are organized as follows:
    R^d -[Linear(d,dh), activation]-> R^dh -[(Linear(dh,dh),activation)*nh]-> R^dh -[Linear(dh,d)]-> R^d
    """
    def __init__(self, *, d, nh, dh, activation_layer_class=NormBiTanh, batch_norm=False):
        super(GeneralBackpropJacobianFlow, self).__init__(d=d)
        self.nh = nh
        self.dh = dh

        assert d <= dh, "The hidden dimension must be larger than the in/out dimension for bijectivity"

        ll = torch.nn.Linear(d, dh, bias=False)
#        torch.nn.init.normal_(ll.weight, std=0.1 / d / nh)
#        ll.weight.data += torch.eye(dh, d).to(ll.weight.data.device)
        layers = [
            ll,
            activation_layer_class(),
        ]
        if batch_norm:
            layers.append(torch.nn.BatchNorm1d(dh))

        for i in range(nh):
            ll = torch.nn.Linear(dh, dh, bias=False)
#            torch.nn.init.normal_(ll.weight, std=0.1 / d / nh)
#            ll.weight.data += torch.eye(dh).to(ll.weight.data.device)
            layers.append(ll)
            layers.append(activation_layer_class())
            if batch_norm:
                layers.append(torch.nn.BatchNorm1d(dh))

        ll = torch.nn.Linear(dh, d, bias=False)
#        torch.nn.init.normal_(ll.weight, std=0.1)
#        ll.weight.data += 5. * torch.eye(d, dh).to(ll.weight.data.device)
        layers.append(ll)
        layers.append(activation_layer_class())

        self.flow_ = (torch.nn.Sequential(*layers))

    def weight_init_identity_(self,std=None):
        """Initialize weights as eye + normal(0,std)
        Essentially realizing a resnet-like layer
        """
        if std is None:
            std_ = 0.1 / self.d / self.nh
        else:
            std_ = std
        for layer in self.flow_:
            if isinstance(layer, torch.nn.Linear):
                torch.nn.init.normal_(layer.weight, std=std_)
                layer.weight.data += torch.eye(self.d).to(layer.weight.data.device)



"""Survey/Refine integrator based on training models with a DKL trainer"""
import torch
from .adaptive_survey_integrator import AdaptiveSurveyIntegrator
from zunis.training.weighted_dataset.dkl_training import BasicStatefulDKLTrainer


class DKLAdaptiveSurveyIntegrator(AdaptiveSurveyIntegrator):
    """Survey/Refine adaptive integrator based on the DKL loss. The loss is the D_KL distance between
    the PDF from a flow model and an un-normalized function, up to non-trainable terms

    Explicitly:

    L(f,q) = - \int dx f(x) log(q(x))

    This integrator is adaptive in the sense that survey batches are sampled from the flat distribution in
    the target space (the domain of f and q) until the learned q distribution is a better approximation of
    the normalized target function f than the flat distribution. Since our target space is the unit hypercube,
    this is easy:

    L(f, uniform) = 0.

    So as soon as the loss is negative, we sample from the flow instead of the uniform distribution.
    """

    def __init__(self, f, trainer, d, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 device=torch.device("cpu"), verbosity=2, trainer_verbosity=1, **kwargs):

        super(DKLAdaptiveSurveyIntegrator, self).__init__(f,
                                                          trainer,
                                                          d,
                                                          n_iter=n_iter,
                                                          n_iter_survey=n_iter_survey,
                                                          n_iter_refine=n_iter_refine,
                                                          n_points=n_points,
                                                          n_points_survey=n_points_survey,
                                                          n_points_refine=n_points_refine,
                                                          use_survey=use_survey,
                                                          device=device,
                                                          verbosity=verbosity,
                                                          trainer_verbosity=trainer_verbosity,
                                                          **kwargs)

    def survey_switch_condition(self):
        """Check if the loss is negative. This test is used to switch from uniform sampling
        to sampling from the flow in the survey phase.

        The loss is the distance between the target function and the flow PDF. Since the distance between
        the target function and the uniform function, a negative loss indicates that flow is doing better.
        """
        return self.model_trainer.record["loss"] < 0



"""Tools to build out-of-the box models"""

from zunis.models.flows.masking import n_ary_mask
from zunis.models.flows.sequential.invertible_sequential import InvertibleSequentialFlow
from zunis.models.flows.coupling_cells.real_nvp import RealNVP
from zunis.models.flows.analytic_flows.element_wise import InvertibleAnalyticSigmoid


def create_n_ary_hypercube_flow(cell, d, n=2, repetitions=2, **cell_options):
    """Create a sequential normalizing flow on the d-dimensional unit hypercube based on a given
    coupling cell with a n-ary mask. Subsequent cells have masks offset by one at each step until
    all possible masks have been used.

    Parameters
    ----------
    cell:
        coupling cell class
    d: int
        dimension of the unit hypercube
    n:
        period of the n-ary mask
    repetitions:
        how many times a full repetition through all possible masks is repeated
    cell_options:
        options to be passed at the coupling cell instantiation

    Returns
    -------
    InvertibleSequentialFlow
    """
    layers = []
    for rep in range(repetitions):
        for i in range(n):
            mask = n_ary_mask(d, n, i)
            layers.append(
                cell(d=d,
                     mask=mask,
                     **cell_options
                     ),
            )
    layers.append(InvertibleAnalyticSigmoid(d=d))

    model = InvertibleSequentialFlow(d, layers)
    return model


def create_n_ary_hypercube_realnvp(d, n=2, repetitions=2, d_hidden=256, n_hidden=16):
    return create_n_ary_hypercube_flow(RealNVP, d, n=n, repetitions=repetitions, d_hidden=d_hidden, n_hidden=n_hidden)


def create_checkerboard_hypercube_realnvp(d, repetitions=2, d_hidden=256, n_hidden=16):
    return create_n_ary_hypercube_realnvp(d, 2, repetitions, d_hidden, n_hidden)


def create_hypercube_flow(d, model_type="realnvp", mask_type="checkerboard", **params):
    """Provide a high-level generator with default behavior"""

    if model_type == "realnvp":
        if mask_type == "checkerboard":
            return create_checkerboard_hypercube_realnvp(d, **params)
        if mask_type == "n_ary":
            return create_n_ary_hypercube_realnvp(d, **params)
        else:
            raise ValueError(f"Mask type {mask_type} does not exist")

    else:
        raise ValueError(f"Model type {model_type} does not exist")



"""Implementation of the piecewise quadratic coupling cell
This means that the *variable transform* is piecewise-quadratic.
"""
import torch
import numpy as np

from ..transforms import InvertibleTransform
from ..general_coupling import InvertibleCouplingCell
from zunis.utils.exceptions import AvertedCUDARuntimeError
from zunis.models.layers.trainable import ArbitraryShapeRectangularDNN
from zunis.models.utils import Reshift

third_dimension_softmax = torch.nn.Softmax(dim=2)

def modified_softmax (v,w):
    v=torch.exp(v)
    vsum=torch.cumsum(v, axis=-1)
    vnorms=torch.cumsum(torch.mul((v[:,:,:-1]+v[:,:,1:])/2,w),axis=-1)
    vnorms_tot=vnorms[:, :, -1].clone() 
    return torch.div(v,torch.unsqueeze(vnorms_tot,axis=-1)) 


def piecewise_quadratic_transform(x, wv_tilde, compute_jacobian=True):
    """Apply an element-wise piecewise-quadratic transformation to some variables

    Parameters
    ----------
    x : torch.Tensor
        a tensor with shape (N,k) where N is the batch dimension while k is the
        dimension of the variable space. This variable spans the k-dimensional unit
        hypercube

    wv_tilde: torch.Tensor
        is a tensor with shape (N,k,2b+1) where b is the number of bins.
        This contains the un-normalized widths and heights of the bins of the piecewise-constant 
        PDF for dimension k,
        i.e. q_tilde lives in all of R and we don't impose a constraint on their sum yet.
        Normalization is imposed in this function using a modified softmax.

    compute_jacobian : bool, optional
        determines whether the jacobian should be compute or None is returned

    Returns
    -------
    tuple of torch.Tensor
        pair `(y,j)`.
        - `y` is a tensor with shape (N,k) living in the k-dimensional unit hypercube
        - `j` is the jacobian of the transformation with shape (N,) if compute_jacobian==True, else None.
    """
    logj = None

    # TODO do a bottom-up assesment of how we handle the differentiability of variables
    
    v_tilde=wv_tilde[:,:,:int(np.ceil(wv_tilde.shape[2]/2))]
    w_tilde=wv_tilde[:,:,v_tilde.shape[2]:]
    N, k, b = w_tilde.shape
    Nx, kx = x.shape
    assert N == Nx and k == kx, "Shape mismatch"
    
    w=torch.exp(w_tilde)
    wsum = torch.cumsum(w, axis=-1) 
    wnorms = torch.unsqueeze(wsum[:, :, -1], axis=-1) 
    w = w/wnorms
    wsum=wsum/wnorms
    wsum_shift=torch.cat((torch.zeros([wsum.shape[0],wsum.shape[1],1]).to(wsum.device, wsum.dtype),wsum),axis=-1)
    
    v=modified_softmax(v_tilde, w)
    
    #tensor of shape (N,k,b+1) with 0 entry if x is smaller than the cumulated w and 1 if it is bigger
    #this is used to find the bin with the number mx in which x lies; for this, the sum of the bin 
    #widths w has to be smaller to the left and bigger to the right
    finder=torch.where(wsum>torch.unsqueeze(x,axis=-1),torch.zeros_like(wsum),torch.ones_like(wsum))
    eps = torch.finfo(wsum.dtype).eps
    #the bin number can be extracted by finding the last index for which finder is nonzero. As wsum 
    #is increasing, this can be found by searching for the maximum entry of finder*wsum. In order to 
    #get the right result when x is in the first bin and finder is everywhere zero, a small first entry 
    #is added
    mx=torch.unsqueeze(  #we need to unsqueeze for later operations
        torch.argmax( #we search for the maximum in order to find the last bin for which x was greater than wsum
            torch.cat((torch.ones([wsum.shape[0],wsum.shape[1],1]).to(wsum.device, wsum.dtype)*eps,finder*wsum),
                      axis=-1),  #we add an offset to ensure that if x is in the first bin, a maximal argument is found 
            axis=-1), 
        axis=-1)
   
    # x is in the mx-th bin: x \in [0,1],
    # mx \in [[0,b-1]], so we clamp away the case x == 1
    mx = torch.clamp(mx, 0, b - 1).to(torch.long)
    # Need special error handling because trying to index with mx
    # if it contains nans will lock the GPU. (device-side assert triggered)
    if torch.any(torch.isnan(mx)).item() or torch.any(mx < 0) or torch.any(mx >= b):
        raise AvertedCUDARuntimeError("NaN detected in PWQuad bin indexing")
    
    # alpha (element of [0.,1], the position of x in its bin)
    # gather collects the cumulated with of all bins until the one in which x lies
    # alpha=(x- Sum_(k=0)^(i-1) w_k)/w_b for x in bin b
    alphas=torch.div((x-torch.squeeze(torch.gather(wsum_shift,-1,mx),axis=-1)),
                         torch.squeeze(torch.gather(w,-1,mx),axis=-1))
    
    #vw_i= (v_i+1 - v_i)w_i/2 where i is the bin index
    vw=torch.cat((torch.zeros([v.shape[0],v.shape[1],1]).to(wsum.device, wsum.dtype),
                                  torch.cumsum(torch.mul((v[:,:,:-1]+v[:,:,1:])/2,w),axis=-1)),axis=-1)
    
    #quadratic term
    out_1=torch.mul((alphas**2)/2,torch.squeeze(torch.mul(torch.gather(v,-1, mx+1)-torch.gather(v,-1, mx),
                                                        torch.gather(w,-1,mx)),axis=-1))
    
    #linear term
    out_2=torch.mul(torch.mul(alphas,torch.squeeze(torch.gather(v,-1,mx),axis=-1)),
                            torch.squeeze(torch.gather(w,-1,mx),axis=-1))
    
    #constant
    out_3= torch.squeeze(torch.gather(vw,-1,mx),axis=-1)
    
    
    out=out_1+out_2+out_3
    
    #the derivative of this transformation is the linear interpolation between v_i-1 and v_i at alpha
    #the jacobian is the product of all linear interpolations
    if compute_jacobian:
        logj=torch.squeeze(
            torch.log(torch.unsqueeze(
                    torch.prod(#we need to take the product over all transformed dimensions
                        torch.lerp(torch.squeeze(torch.gather(v,-1,mx),axis=-1),
                                   torch.squeeze(torch.gather(v,-1,mx+1),axis=-1),alphas),
                        #linear extrapolation between alpha, mx and mx+1
                    axis=-1),
                axis=-1)),
            axis=-1)
       
    # Regularization: points must be strictly within the unit hypercube
    # Use the dtype information from pytorch
    eps = torch.finfo(out.dtype).eps
    out = out.clamp(
        min=eps,
        max=1. - eps
    )
    return out, logj


def piecewise_quadratic_inverse_transform(y, wv_tilde, compute_jacobian=True):
    """
    Apply the inverse of an element-wise piecewise-linear transformation to some variables

    Parameters
    ----------
    y : torch.Tensor
        a tensor with shape (N,k) where N is the batch dimension while k is the
        dimension of the variable space. This variable span the k-dimensional unit
        hypercube

    wv_tilde: torch.Tensor
        is a tensor with shape (N,k,2b+1) where b is the number of bins.
        This contains the un-normalized widths and heights of the bins of the piecewise-constant PDF for dimension k,
        i.e. q_tilde lives in all of R and we don't impose a constraint on their sum yet.
        Normalization is imposed in this function using a modified softmax.
        

    compute_jacobian : bool, optional
        determines whether the jacobian should be compute or None is returned

    Returns
    -------
    tuple of torch.Tensor
        pair `(x,j)`.
        - `x` is a tensor with shape (N,k) living in the k-dimensional unit hypercube
        - `j` is the jacobian of the transformation with shape (N,) if compute_jacobian==True, else None.
    """
    
    logj = None

    # TODO do a bottom-up assesment of how we handle the differentiability of variables
    
    v_tilde=wv_tilde[:,:,:int(np.ceil(wv_tilde.shape[2]/2))]
    w_tilde=wv_tilde[:,:,v_tilde.shape[2]:]
    N, k, b = w_tilde.shape
    
    Nx, kx = y.shape
    assert N == Nx and k == kx, "Shape mismatch"
    
    w=torch.exp(w_tilde)
    wsum = torch.cumsum(w, axis=-1) 
    wnorms = torch.unsqueeze(wsum[:, :, -1], axis=-1) 
    w = w/wnorms
    wsum=wsum/wnorms
    wsum_shift=torch.cat((torch.zeros([wsum.shape[0],wsum.shape[1],1]).to(wsum.device, wsum.dtype),wsum),axis=-1)
    
    v=modified_softmax(v_tilde, w)
    
    #need to find the bin number for each of the y/x
    #-> find the last bin such that y is greater than the constant of the quadratic equation
    
    #vw_i= (v_i+1 - v_i)w_i/2 where i is the bin index. VW is the constant of the quadratic equation
    vw=torch.cat((torch.zeros([v.shape[0],v.shape[1],1]).to(wsum.device, wsum.dtype),
                                  torch.cumsum(torch.mul((v[:,:,:-1]+v[:,:,1:])/2,w),axis=-1)),axis=-1)
    # finder is contains 1 where y is smaller then the constant and 0 if it is greater
    finder=torch.where(vw>torch.unsqueeze(y,axis=-1),torch.zeros_like(vw),torch.ones_like(vw))
    eps = torch.finfo(vw.dtype).eps
    #the bin number can be extracted by finding the last index for which finder is nonzero. As vw 
    #is increasing, this can be found by searching for the maximum entry of finder*vw. In order to 
    #get the right result when y is in the first bin and finder is everywhere zero, a small first entry 
    #is added and mx is reduced by one to account for the shift.
    mx=torch.unsqueeze(
        torch.argmax(#we search for the maximum in order to find the last bin for which y was greater than vw
            torch.cat((torch.ones([vw.shape[0],vw.shape[1],1]).to(vw.device, vw.dtype)*eps,finder*(vw+1)),axis=-1),
            axis=-1), #we add an offset to ensure that if x is in the first bin, a maximal argument is found
        axis=-1)-1 # we substract -1 to account for the offset
    
    # x is in the mx-th bin: x \in [0,1],
    # mx \in [[0,b-1]], so we clamp away the case x == 1
    edges = torch.clamp(mx, 0, b - 1).to(torch.long)
    
    # Need special error handling because trying to index with mx
    # if it contains nans will lock the GPU. (device-side assert triggered)
    if torch.any(torch.isnan(edges)).item() or torch.any(edges < 0) or torch.any(edges >= b):
        raise AvertedCUDARuntimeError("NaN detected in PWQuad bin indexing")
    
    #solve quadratic equation
    
    #prefactor of quadratic term
    a=torch.squeeze(torch.mul(torch.gather(v,-1, edges+1)-torch.gather(v,-1, edges),
                                                        torch.gather(w,-1,edges)),axis=-1)
    #prefactor of linear term
    b=torch.mul(torch.squeeze(torch.gather(v,-1,edges),axis=-1),torch.squeeze(torch.gather(w,-1,edges),axis=-1))
    #constant - y
    c= torch.squeeze(torch.gather(vw,-1,edges),axis=-1)-y
    
    #ensure that division by zero is taken care of
    eps = torch.finfo(a.dtype).eps
    a=torch.where(torch.abs(a)<eps,eps*torch.ones_like(a),a)
    
    d = (b**2) - (2*a*c)
    
    assert not torch.any(d<0), "Value error in PWQuad inversion"
    assert not torch.any(a==0), "Value error in PWQuad inversion, a==0"
    
    # find two solutions
    sol1 = (-b-torch.sqrt(d))/(a)
    sol2 = (-b+torch.sqrt(d))/(a)
    
    # choose solution which is in the allowed range
   
    sol=torch.where((sol1>=0)&(sol1<1), sol1, sol2)
    
    if torch.any(torch.isnan(sol)).item():
        raise AvertedCUDARuntimeError("NaN detected in PWQuad inversion")
    
    eps = torch.finfo(sol.dtype).eps
    
    
    sol = sol.clamp(
        min=eps,
        max=1. - eps
    )
    
    #the solution is the relative position inside the bin. This can be
    #converted to the absolute position by adding the sum of the bin widths up to this bin
    x=torch.mul(torch.squeeze(torch.gather(w,-1,edges),axis=-1), sol)+torch.squeeze(torch.gather(wsum_shift,-1,edges),axis=-1)
    
    eps = torch.finfo(x.dtype).eps
    
    x = x.clamp(
        min=eps,
        max=1. - eps
    )
    
    if compute_jacobian:
        logj =-torch.squeeze(torch.log(
            torch.unsqueeze(torch.prod(#we have to take the product of the jacobian of all dimensions
                torch.lerp(torch.squeeze(torch.gather(v,-1,edges),axis=-1),torch.squeeze(torch.gather(v,-1,edges+1),
                                                                                         axis=-1),sol),
                axis=-1), #linear extrapolation between sol, edges and edges+1 gives the jacobian of the forward transformation. The prefactor of -1 is the log of the jacobian of the inverse
            axis=-1)),
        axis=-1)
       
    return x.detach(), logj


class ElementWisePWQuadraticTransform(InvertibleTransform):
    """Invertible piecewise-quadratic transformations over the unit hypercube

    Implements a batched bijective transformation `h` from the d-dimensional unit hypercube to itself,
    in an element-wise fashion (each coordinate transformed independently)

    In each direction, the bijection is a piecewise-quadratic transform with b bins
    where the forward transform has bins with adjustable width. The transformation in each bin is
    then a quadratic spline. The network predicts the bin width w_tilde and the vertex height v_tilde of the
    derivative of the transform for each direction and each point of the batch. They are normalized such that: 
    1. h(0) = 0
    2. h(1) = 1
    3. h is monotonous
    4. h is continuous

    Conditions 1. to 3. ensure the transformation is a bijection and therefore invertible.
    The inverse is also an element-wise, piece-wise quadratic transformation.
    """

    backward = staticmethod(piecewise_quadratic_transform)
    forward = staticmethod(piecewise_quadratic_inverse_transform)


class GeneralPWQuadraticCoupling(InvertibleCouplingCell):
    """Abstract class implementing a coupling cell based on PW quadratic transformations

    A specific way to predict the parameters of the transform must be implemented
    in child classes.
    """

    def __init__(self, *, d, mask):
        """Generator for the abstract class GeneralPWQuadraticCoupling

        Parameters
        ----------
        d: int
            dimension of the space
        mask: list of bool
            variable mask which variables are transformed (False)
            or used as parameters of the transform (True)

        """
        super(GeneralPWQuadraticCoupling, self).__init__(d=d, transform=ElementWisePWQuadraticTransform(), mask=mask)


class PWQuadraticCoupling(GeneralPWQuadraticCoupling):
    """Piece-wise Quadratic coupling

    Coupling cell using an element-wise piece-wise quadratic transformation as a change of
    variables. The transverse neural network is a rectangular dense neural network.

    Notes:
        Transformation used:
        `zunis.models.flows.coupling_cells.piecewise_coupling.piecewise_quadratic.ElementWisePWQuadraticTransform`
        Neural network used:
        zunis.models.layers.trainable.ArbitraryShapeRectangularDNN
    """

    def __init__(self, *, d, mask,
                 n_bins=10,
                 d_hidden=256,
                 n_hidden=8,
                 input_activation=Reshift,
                 hidden_activation=torch.nn.LeakyReLU,
                 output_activation=None,
                 use_batch_norm=False):
        """
        Generator for PWQuadraticCoupling

        Parameters
        ----------
        d: int
        mask: list of bool
            variable mask: which dimension are transformed (False) and which are not (True)
        n_bins: int
            number of bins in each dimensions
        d_hidden: int
            dimension of the hidden layers of the DNN
        n_hidden: int
            number of hidden layers in the DNN
        input_activation: optional
            pytorch activation function before feeding into the DNN.
            must be a callable generator without arguments (i.e. a classname or a function)
        hidden_activation: optional
            pytorch activation function between hidden layers of the DNN.
            must be a callable generator without arguments (i.e. a classname or a function)
        output_activation: optional
            pytorch activation function at the output of the DNN.
            must be a callable generator without arguments (i.e. a classname or a function)
        use_batch_norm: bool
            whether batch normalization should be used in the DNN.
        """

        super(PWQuadraticCoupling, self).__init__(d=d, mask=mask)

        d_in = sum(mask)
        d_out = d - d_in

        self.T = ArbitraryShapeRectangularDNN(d_in=d_in,
                                              out_shape=(d_out, 2*n_bins+1),
                                              d_hidden=d_hidden,
                                              n_hidden=n_hidden,
                                              input_activation=input_activation,
                                              hidden_activation=hidden_activation,
                                              output_activation=output_activation,
                                              use_batch_norm=use_batch_norm)



"""Main weighted dataset stateful trainer API class"""
from functools import partial
import torch

from .weighted_dataset_trainer import BasicStatefulTrainer
from zunis.models.flows.sequential.repeated_cell import RepeatedCellFlow
from .dkl_training import weighted_dkl_loss
from .variance_training import weighted_variance_loss
from zunis.models.flows.sampling import UniformSampler, FactorizedGaussianSampler, FactorizedFlowSampler


class StatefulTrainer(BasicStatefulTrainer):
    """High-level API for stateful trainers using weighted datasets
    (dataset consisting of tuples of point, function value, point pdf).
    """

    loss_map = {
        "dkl": weighted_dkl_loss,
        "variance": weighted_variance_loss
    }
    """Dictionary for the string-based API to define the loss function used in training"""

    default_flow_priors = {
        "realnvp": FactorizedGaussianSampler,
        "pwlinear": UniformSampler,
        "pwquad": UniformSampler}
    """Dictionary for the string-based API to define the distribution of the data in latent space based on
    the choice of coupling cell"""

    flow_priors = {
        "gaussian": FactorizedGaussianSampler,
        "uniform": UniformSampler
    }
    """Dictionary for the string-based API to define the distribution of the data in latent space"""

    def __init__(self, d, loss="variance", flow="pwlinear", flow_prior=None, flow_options=None, prior_options=None,
                 device=torch.device("cpu"), n_epochs=10, optim=None, **kwargs):
        """

        Parameters
        ----------
        d: int
            dimensionality of the space
        loss: str or function
            loss function. If this argument is a string, it is mapped to a function using
            :py:attr:`zunis.training.weighted_dataset.stateful_trainer.loss_map`
        flow: str or :py:class:`zunis.models.flows.general_flow.GeneralFlow`
            if this variable is a string, it is a cell key used in :py:class:`zunis.models.flows.sequential.repeated_cell.RepeatedCellFlow`
            otherwise it can be an actual flow model
        flow_prior: None or str or :py:class:`zunis.models.flows.sampling.FactorizedFlowSampler`
            PDF used for sampling latent space. If None (default) then use the "natural choice" defined
            in the class variable :py:attr:`zunis.training.weighted_dataset.stateful_trainer.StatefulTrainer.default_flow_priors`
            A string argument will be mapped using :py:attr:`zunis.training.weighted_dataset.stateful_trainer.StatefulTrainer.flow_priors`
        flow_options: None or dict
            options to be passed to the :py:class:`zunis.models.flows.sequential.repeated_cell.RepeatedCellFlow` model if
            `flow` is a string
        prior_options: None or dict
            options to be passed to the latent prior constructor if a "natural choice" prior is used
            i.e. if `flow_prior` is `None` or a `str`
        device:
            device on which to run the model and the sampling
        n_epochs: int
            number of epochs per batch of data during training
        optim: None or torch.optim.Optimizer sublcass
            optimizer to use for training. If none, default Adam is used
        """

        # The loss function is either a string that points to the loss_map class dictionary or an actual loss function
        if isinstance(loss, str):
            loss = self.loss_map[loss]

        # The prior has three options:
        #
        # - None: if we use the string-based interface to define a RepeatedCellFlow, use the associated default prior
        # in `self.default_flow_priors`
        #
        # - string: use `self.flow_priors`
        #
        # - an actual prior object to be used as-is
        if prior_options is None:
            prior_options = dict()

        if flow_prior is None:
            assert isinstance(flow,
                              str), "You must specify a flow if you are not using the string-based model interface"
            flow_prior = self.default_flow_priors[flow](d=d, device=device, **prior_options)
        elif isinstance(flow_prior, str):
            flow_prior = self.flow_priors[flow_prior](d=d, device=device, **prior_options)
        else:
            assert isinstance(flow_prior,
                              FactorizedFlowSampler), \
                "The flow prior must be either None, a string or a FactorizedFlowSampler"

        # Two options for flows: either use the RepeatedCellFlow interface or pass an actual flow object
        if isinstance(flow, str):
            if flow_options is None:
                flow_options = dict()
            flow = RepeatedCellFlow(d=d, cell=flow, **flow_options).to(device)

        if optim is None:
            optim_ = torch.optim.Adam(flow.parameters())
        else:
            optim_ = optim(flow.parameters())

        super(StatefulTrainer, self).__init__(flow, flow_prior, n_epochs=n_epochs, optim=optim_, **kwargs)
        self.loss = loss



"""Optimization of invertible flows in the weighted dataset problem using the DKL loss

Reminder: we have a dataset (x,p(x),f(x)) such that
- x ~ p(x)
- we want to learn a model that draws points according to f(x),
which is positive, and known up to normalization

We can estimate the expected un-normalized log-likelihood of a random datapoint x~f(x)
under the pdf of our flow model q(x) as

lL(f,q) = ∫ dx f(x) łog(q(x)) = E(f(x)/p(x) log(q(x)), x~p(x))

This is the same as minimizing the DKL between φ(x)=f(x)/∫ dx f(x) and q(x):

D_{KL}(φ|q) = ∫ dx φ(x) łog(φ(x)/q(x)) = - lL(f,q)*λ + η

where λ>0 and η are constants that are independent of the parameters of our flow q so maximizing the
estimated log-likelihood is the same as minimizing the DKL. In either case, we can only do it up to a constant
and use the loss lL(f,q) as a proxy

NB: An important point for importance sampling.
In importance sampling, a natural loss is the variance of the integrand. One can however note that
the optimum for both the variance loss and the DKL/ML is when the flow reproduces the target distribution.
"""
import logging
import torch
from .weighted_dataset_trainer import BasicTrainer, BasicStatefulTrainer

logger = logging.getLogger(__name__)


def weighted_dkl_loss(fx, px, logqx):
    return - torch.mean(fx * logqx / px)


class BasicDKLTrainer(BasicTrainer):
    def __init__(self, flow, latent_prior):
        super(BasicDKLTrainer, self).__init__(flow, latent_prior)
        self.loss = weighted_dkl_loss


class BasicStatefulDKLTrainer(BasicStatefulTrainer):
    def __init__(self, flow, latent_prior, **kwargs):
        super(BasicStatefulDKLTrainer, self).__init__(flow, latent_prior, **kwargs)
        self.loss = weighted_dkl_loss



"""Module for the generation of masks for coupling cells"""


def get_bin(x, n=0):
    """
    Get the binary representation of x.

    Parameters
    ----------
    x : int
    n : int
        Minimum number of digits. If x needs less digits in binary, the rest
        is filled with zeros.

    Returns
    -------
    list of binary digits
    """

    y = format(x, 'b').zfill(n)

    return [int(i) for i in str(y)]


def n_ary_mask(d, n, offset):
    """Create a n-ary mask with n entries (a list of bool with each nth entry True)

    Parameters
    ----------
    d: int
        numbers of entries in the mask
    n: int
        period of the mask
    offset: int
        offset of the True entries

    Returns
    -------
    list of bool
        True/False mask with each nth entry True
    """
    return [(i + offset) % n == 0 for i in range(d)]


def n_ary_mask_strategy(d, n=2, repetitions=1):
    """Generate an list of masks using the n-ary periodic strategy: every nth entry is True and each next mask is
    offset by one position with respect to the previous one. We loop for a given number of repetitions through all
    possible masks.

    Parameters
    ----------
    d: int
        dimensionality of the space
    n: int
        periodicity of the mask
    repetitions: int
        number of iterations through all possible masks

    Returns
    -------
    list of list of bool
        list of masks
    """

    masks = []
    for rep in range(repetitions):
        for i in range(n):
            masks.append(n_ary_mask(d, n, i))

    return masks


def iflow_strategy(d, repetitions=1):
    """Generate a list of masks using the strategy of Gao et al. arXiv:2001.05486.
    Each dimension is numbered and the number is converted into the binary representation. Then,
    one bit after another, starting with the most significant bit, is converted into a boolean.
    This determines wether the dimension is transformed in this cell or not. In order to be 
    able to test for all correlations, for each mask, the inverse mask is used too. 
    

    The minimal amount of cells will be used to represent all possible correlations.

    Parameters
    ----------
    d: int
        number of dimensions

    Returns
    -------
    list of list of bool
        list of masks

    """

    n = len(get_bin(d - 1, 0))
    masks = [1] * 2 * n * repetitions
    dims = [int(i) for i in range(d)]

    dims_bin = [list(x) for x in zip(*list(map(get_bin, dims, [n] * d)))]
    j = 0
    for k in range(repetitions):
        for i in range(n):
            s = dims_bin[i][:]

            masks[j] = [bool(x) for x in s]

            masks[j + 1] = [not bool(x) for x in s]
            j = j + 2

    return (masks)


def maximal_masking_strategy(d, repetitions=1):
    """Generate a list of masks using the maximally complex option:
    (not really but within the reasonable options)

    Each variable is transformed using all the others as parameters, looping
    over all variables. The size of the model is therefore linear with the
    number of dimensions

    Parameters
    ----------
    d: int
        number of dimensions
    repetitions: int
        number of repetitions

    Returns
    -------

    """

    masks = [[False] * d for _ in range(d * repetitions)]
    for r in range(repetitions):
        for i in range(d):
            masks[i + r * d][i] = True
    return masks


"""Trainable layers
"""
import torch

class OverallAffineLayer(torch.nn.Module):
    """Learnable overall affine transformation
    f(x) = alpha x + delta
    """

    def __init__(self, alpha=10., delta=0.):
        super().__init__()
        self.alpha = torch.nn.Parameter(torch.tensor(alpha), requires_grad=True)
        self.delta = torch.nn.Parameter(torch.tensor(delta), requires_grad=True)

    def forward(self, input):
        """Output of the OverallAffineLayer"""
        return input * self.alpha + self.delta


def create_rectangular_dnn(
        *,
        d_in,
        d_out,
        d_hidden,
        n_hidden,
        input_activation=None,
        hidden_activation=torch.nn.ReLU,
        output_activation=None,
        use_batch_norm=False):
        
        layers = []
        if input_activation is not None:
            layers.append(input_activation())
        layers.append(torch.nn.Linear(d_in,d_hidden))
        layers.append(hidden_activation())
        if use_batch_norm:
            layers.append(torch.nn.BatchNorm1d(d_hidden))

        for i in range(n_hidden):
            layers.append(torch.nn.Linear(d_hidden, d_hidden))
            layers.append(hidden_activation())
            if use_batch_norm:
                layers.append(torch.nn.BatchNorm1d(d_hidden))

        layers.append(torch.nn.Linear(d_hidden, d_out))

        if output_activation is not None:
            layers.append(output_activation())

        return torch.nn.Sequential(*layers)


class ArbitraryShapeRectangularDNN(torch.nn.Module):
    """Rectangular DNN with the output layer reshaped to a given shape"""
    def __init__(self, *,
                 d_in,
                 out_shape,
                 d_hidden,
                 n_hidden,
                 input_activation=None,
                 hidden_activation=torch.nn.ReLU,
                 output_activation=None,
                 use_batch_norm=False):

        super(ArbitraryShapeRectangularDNN, self).__init__()
        self.out_shape = out_shape

        d_out = 1
        for d in out_shape:
            d_out *= d

        self.nn = create_rectangular_dnn(d_in=d_in,
                                         d_out=d_out,
                                         d_hidden=d_hidden,
                                         n_hidden=n_hidden,
                                         input_activation=input_activation,
                                         hidden_activation=hidden_activation,
                                         output_activation=output_activation,
                                         use_batch_norm=use_batch_norm)

    def forward(self, x):
        return self.nn(x).view(*(x.shape[:-1]), *self.out_shape)




from functools import partial

import torch

from .invertible_sequential import InvertibleSequentialFlow
from zunis.models.flows.coupling_cells.general_coupling import InvertibleCouplingCell
from zunis.models.flows.analytic_flows.analytic_flow import InvertibleAnalyticFlow
from ..masking import n_ary_mask_strategy, maximal_masking_strategy,iflow_strategy

from zunis.models.flows.coupling_cells.real_nvp import RealNVP
from zunis.models.flows.coupling_cells.piecewise_coupling.piecewise_linear import PWLinearCoupling
from zunis.models.flows.coupling_cells.piecewise_coupling.piecewise_quadratic import PWQuadraticCoupling
from zunis.models.flows.analytic_flows.element_wise import InvertibleAnalyticSigmoid


class MaskListRepeatedCellFlow(InvertibleSequentialFlow):
    """Sequential normalizing flow based on repeating a single coupling cell type
    with each a different mask as a sequence of transformations"""

    def __init__(self, d, cell, masks, *, input_cell=None, output_cell=None, cell_params=None, input_cell_params=None,
                 output_cell_params=None):
        """

        Parameters
        ----------
        d: int
            dimensionality of the space
        cell:
            callable that yields an instance of zunis.models.flows.coupling_cells.general_coupling.InvertibleCouplingCell
            the coupling cell to be repeated.
        masks: list of lists of bool
            list of masks for each flow in the cell. The order is from latent space to target space
        input_flow: , None
            flow model to be used as an input. Typically to deform the input space to be suited to the repeated cell
        output_flow: , None
            flow model to be used as an output. Typically to deform the input space to be suited to the repeated cell
        cell_params: dict
            parameters of the cell
        input_cell_params: dict, None
            parameters of the input cell
        output_cell_params: dict
            parameters of the output cell
        """
        assert issubclass(cell, InvertibleCouplingCell), "The repeated cell must be an InvertibleCouplingCell"

        if cell_params is None:
            cell_params = dict()
        if input_cell_params is None:
            input_cell_params = dict()
        if output_cell_params is None:
            output_cell_params = dict()

        flows = []
        if input_cell is not None:
            assert issubclass(input_cell, InvertibleAnalyticFlow), "The input cell must be an InvertibleAnalyticFlow"
            flows.append(input_cell(d=d, **input_cell_params))

        for mask in masks:
            flows.append(cell(d=d, mask=mask, **cell_params))

        if output_cell is not None:
            assert issubclass(output_cell, InvertibleAnalyticFlow), "The output cell must be an InvertibleAnalyticFlow"
            flows.append(output_cell(d=d, **output_cell_params))

        super(MaskListRepeatedCellFlow, self).__init__(d=d, flows=flows)


class RepeatedCellFlow(MaskListRepeatedCellFlow):
    """Repeated cell flow: a sequential model with a coupling cell model
    This is meant as a higher-level API to generate boilerplate models based on architecture and masking strategy
    """
    cells = {"realnvp": (RealNVP, None, InvertibleAnalyticSigmoid),
             "pwlinear": (PWLinearCoupling, None, None),
             "pwquad": (PWQuadraticCoupling, None, None)}

    masking = {
        "checkerboard": partial(n_ary_mask_strategy, n=2),
        "maximal": maximal_masking_strategy,
        "iflow": iflow_strategy
    }

    def __init__(self, d, cell="pwlinear", masking="checkerboard", *, input_cell=None, output_cell=None,
                 cell_params=None, input_cell_params=None, output_cell_params=None, masking_options=None):
        """

        Parameters
        ----------
        d: int
            dimensionality
        cell: {"pwlinear", "pwquad", "realnvp"}
            coupling cell choice
        masking: {"checkerboard", "maximal"}
            masking strategy. See :py:attr:`RepeatedCellFlow.masking <zunis.models.flows.sequential.repeated_cell.RepeatedCellFlow.masking>`
            and :py:mod:`masking <zunis.models.flows.masking>`
        input_cell: optional
            Transformation before the repeated cell
        output_cell: optional
            Transformation after the repeated cell
        cell_params: dict, optional
            parameters for the repeated cell
        input_cell_params: dict, optional
            parameters for the input_cell
        output_cell_params: dict, optional
            parameters for the output_cell
        masking_options: dict, optional
            parameters for the masking strategy
        """

        if isinstance(cell, str):
            cell, input_cell_, output_cell_ = self.cells[cell]
        else:
            input_cell_ = None
            output_cell_ = None

        if input_cell is None:
            input_cell = input_cell_
        if output_cell is None:
            output_cell = output_cell_
        if masking_options is None:
            masking_options = dict()

        masks = self.masking[masking](d, **masking_options)
       
        super(RepeatedCellFlow, self).__init__(d, cell=cell, masks=masks,
                                               input_cell=input_cell, output_cell=output_cell,
                                               cell_params=cell_params, input_cell_params=input_cell_params,
                                               output_cell_params=output_cell_params)



"""Abstract definition of invertible transforms for coupling cells"""
from better_abc import ABC,abstractmethod
import torch


class InvertibleTransform(ABC):
    """Callable invertible transform which calls self.forward or self.backward depending on self.inverse

    Attributes
    ----------
    inverse: bool
        whether calling the object will use the forward or backward transform
    """
    def __init__(self):
        """Constructor for InvertibleTransform

        No arguments, no returns, this is just encapsulation for a function and its inverse
        """
        self.inverse = False

    @abstractmethod
    def forward(self, y, T, compute_jacobian=True):
        """Abstract forward transform

        Parameters
        ----------
        y: torch.Tensor
            batch of points to transform
        T: torch.Tensor
            batch of parameters for the transformation
        compute_jacobian: bool
            whether to compute the jacobian or not

        Returns
        -------
        tuple of torch.Tensor, torch.Tensor
            `(z,j)` where the first tensor is the batch of transformed points and the second is the batch of jacobians
        """
        pass

    @abstractmethod
    def backward(self, y, T, compute_jacobian=True):
        """Abstract backward transform

        Parameters
        ----------
        y: torch.Tensor
            batch of points to transform
        T: torch.Tensor
            batch of parameters for the transformation
        compute_jacobian: bool
            whether to compute the jacobian or not

        Returns
        -------
        tuple of torch.Tensor, torch.Tensor
            `(z,j)` where the first tensor is the batch of transformed points and the second is the batch of jacobians
        """
        pass

    def __call__(self,y , T, compute_jacobian=True):
        """A transform is callable and either performs the forward or inverse transformation
        depending on the attribute `self.inverse`.

        Parameters
        ----------
        y: torch.Tensor
            batch of points to transform
        T: torch.Tensor
            batch of parameters for the transformation
        compute_jacobian: bool
            whether to compute the jacobian or not


        Returns
        -------
        tuple of torch.Tensor, torch.Tensor
            `(z,j)` where the first tensor is the batch of transformed points and the sec
        """
        if self.inverse:
            return self.backward(y, T)
        else:
            return self.forward(y, T)

    def invert(self):
        """Flip the `inverse` attribute

        Returns
        -------
        None
        """
        self.inverse = not self.inverse




"""Custom exceptions for the ZüNIS library"""


class AvertedCUDARuntimeError(RuntimeError):
    """Prevents the raising of a RuntimeError cause by a CUDA device-side assert failure
    which whould lock the GPU in a failed state."""


class NoCheckpoint(RuntimeError):
    """Raised to signify no checkpoint is available for reloading"""


class TrainingInterruption(RuntimeError):
    """Raised when a Trainer determines that the model should no longer be trained"""



"""Integrator that does not sample points during the training phase but uses a fixed dataset of points"""
from collections.abc import Sequence, Mapping
import pickle

import numpy as np
import torch

from .base_integrator import BaseIntegrator


class FixedSampleSurveyIntegrator(BaseIntegrator):
    """Integrator that trains its model during the survey phase using a pre-computed sample provided externally"""

    def __init__(self, f, trainer, sample=None, n_iter=None, n_iter_survey=1, n_iter_refine=10,
                 n_points=None, n_points_survey=None, n_points_refine=10000, use_survey=False,
                 verbosity=None, trainer_verbosity=None, **kwargs):
        """

        Parameters
        ----------
        f: callable
            ZuNIS-compatible function
        trainer: BasicTrainer
            trainer object used to perform the survey
        sample: tuple of torch.Tensor
            (x, fx, px): target-space point batch drawn from some PDF p, function value batch, PDF value batch p(x)
        n_iter: int
            number of iterations (used for both survey and  refine unless specified)
        n_iter_survey: int
            number of iterations for survey
        n_iter_refine: int
            number of iterations for refine
        n_points: int
            number of points for both survey and refine unless specified
        n_points_survey: int
            number of points for survey
        n_points_refine: int
            number of points for refine
        use_survey: bool
            whether to use the integral estimations from the survey phase. This makes error estimation formally
            incorrect since samples from the refine depend on the survey training, but these correlation can be negligible
            in some cases.
        verbosity: int
            level of verbosity for the integrator-level logger
        trainer_verbosity: int
            level of verbosity for the trainer-level logger
        kwargs
        """
        super(FixedSampleSurveyIntegrator, self).__init__(f=f,
                                                          trainer=trainer,
                                                          trainer_verbosity=trainer_verbosity,
                                                          n_iter=n_iter,
                                                          n_iter_survey=n_iter_survey,
                                                          n_iter_refine=n_iter_refine,
                                                          n_points=n_points,
                                                          n_points_survey=n_points_survey,
                                                          n_points_refine=n_points_refine,
                                                          use_survey=use_survey,
                                                          verbosity=verbosity,
                                                          **kwargs)

        self.sample = sample

    def sample_survey(self, n_points=None, **kwargs):
        """Sample points from the internally stored sample

        Parameters
        ----------
        n_points: int, None
            size of the batch to select from the sample
        kwargs

        Returns
        -------
            tuple of torch.Tensor
                (x,px,fx): sampled points, sampling distribution PDF values, function values

        """
        assert self.sample is not None, "The training sample must be instantiated before starting the survey"

        if n_points is None:
            n_points = self.n_points_survey

        if n_points is None:
            return self.sample

        x, px, fx = self.sample

        sample_size = x.shape[0]
        assert n_points <= sample_size

        # Sample n_points indices randomly among the sample_size options with equal probability
        indices = torch.multinomial(torch.ones(sample_size), n_points)
        return x[indices], px[indices], fx[indices]

    def set_sample(self, sample):
        """Assign a sample to be trained on

        Parameters
        ----------
        sample: tuple of torch.Tensor
            (x,px,fx): sampled points, sampling distribution PDF values, function values
        """
        assert len(sample) == 3, "A sample is a Sequence of three elements"
        for el in sample:
            assert isinstance(el, torch.Tensor), "All elements of a sample must be tensors"
        assert sample[0].shape[0] == sample[1].shape[0] == sample[2].shape[0], \
            "All elements of a sample must share the same batch size"
        assert len(sample[0].shape) == 2 and len(sample[1].shape) == len(sample[2].shape) == 1, \
            "Sample shapes must be (n_batch, n_dim), (n_batch), (n_batch)"
        assert sample[0].shape[1] == self.model_trainer.flow.d, "Number of dimensions must match the flow"

        self.sample = sample

    def set_sample_pickle(self, pickle_path, device=None):
        """Assign a sample to be trained on from a pickle file
        The pickle file must either contain a tuple (x,px,fx) of point batch, PDF value batch, function batch
        or a mapping with keys "x", "px", "fx". In either case, these batches must be valid inputs for torch.tensor

        Parameters
        ----------
        pickle_path: str
            path to the pickle file.
        device: torch.device, None
            device on which to send the sample. If none is provided, flow parameter device will be used
        """
        with open(pickle_path, "rb") as picklefile:
            pickle_sample = pickle.load(picklefile)

        if device is None:
            device = next(self.model_trainer.flow.parameters()).device

        if isinstance(pickle_sample, Sequence):
            self.set_sample([torch.tensor(el).to(device) for el in pickle_sample])

        elif isinstance(pickle_sample, Mapping):
            self.set_sample([
                torch.tensor(pickle_sample[key]).to(device) for key in ["x", "px", "fx"]
            ])

        else:
            raise TypeError("Pickled sample must be either sequences or mappings")

    def set_sample_csv(self, csv_path, device=None, delimiter=",", dtype=np.float):
        """Assign a sample to be trained on from a csv file
        The file must contain equal length rows with at least four columns, all numerical.
        All columns but the last two are interpreted as point coordinates,
        the next-to-last is the point PDF and the last is the function value.

        Parameters
        ----------
        csv_path: str
            path to the csv file
        device: torch.device
            device to which to send the sample
        delimiter: str
            delimiter of the csv file
        """

        data = np.genfromtxt(csv_path, delimiter=delimiter, dtype=dtype)
        x = torch.tensor(data[:, :-2]).to(device)
        px = torch.tensor(data[:, -2]).to(device)
        fx = torch.tensor(data[:, -1]).to(device)

        self.set_sample((x, px, fx))



"""Implementation of the piecewise linear coupling cell
This means that the *variable transform* is piecewise-linear.
"""
import torch

from ..transforms import InvertibleTransform
from ..general_coupling import InvertibleCouplingCell
from zunis.utils.exceptions import AvertedCUDARuntimeError
from zunis.models.layers.trainable import ArbitraryShapeRectangularDNN
from zunis.models.utils import Reshift

third_dimension_softmax = torch.nn.Softmax(dim=2)


def piecewise_linear_transform(x, q_tilde, compute_jacobian=True):
    """Apply an element-wise piecewise-linear transformation to some variables

    Parameters
    ----------
    x : torch.Tensor
        a tensor with shape (N,k) where N is the batch dimension while k is the
        dimension of the variable space. This variable span the k-dimensional unit
        hypercube

    q_tilde: torch.Tensor
        is a tensor with shape (N,k,b) where b is the number of bins.
        This contains the un-normalized heights of the bins of the piecewise-constant PDF for dimension k,
        i.e. q_tilde lives in all of R and we don't impose a constraint on their sum yet.
        Normalization is imposed in this function using softmax.

    compute_jacobian : bool, optional
        determines whether the jacobian should be compute or None is returned

    Returns
    -------
    tuple of torch.Tensor
        pair `(y,h)`.
        - `y` is a tensor with shape (N,k) living in the k-dimensional unit hypercube
        - `j` is the jacobian of the transformation with shape (N,) if compute_jacobian==True, else None.
    """

    logj = None

    # TODO do a bottom-up assesment of how we handle the differentiability of variables

    # Compute the bin width w
    N, k, b = q_tilde.shape
    Nx, kx = x.shape
    assert N == Nx and k == kx, "Shape mismatch"

    w = 1. / b

    # Compute the normalized bin heights by applying a softmax function on the bin dimension
    q = 1. / w * third_dimension_softmax(q_tilde)

    # x is in the mx-th bin: x \in [0,1],
    # mx \in [[0,b-1]], so we clamp away the case x == 1
    mx = torch.clamp(torch.floor(b * x), 0, b - 1).to(torch.long)
    # Need special error handling because trying to index with mx
    # if it contains nans will lock the GPU. (device-side assert triggered)
    if torch.any(torch.isnan(mx)).item() or torch.any(mx < 0) or torch.any(mx >= b):
        raise AvertedCUDARuntimeError("NaN detected in PWLinear bin indexing")

    # We compute the output variable in-place
    out = x - mx * w  # alpha (element of [0.,w], the position of x in its bin

    # Multiply by the slope
    # q has shape (N,k,b), mxu = mx.unsqueeze(-1) has shape (N,k) with entries that are a b-index
    # gather defines slope[i, j, k] = q[i, j, mxu[i, j, k]] with k taking only 0 as a value
    # i.e. we say slope[i, j] = q[i, j, mx [i, j]]
    slopes = torch.gather(q, 2, mx.unsqueeze(-1)).squeeze(-1)
    out = out * slopes
    # The jacobian is the product of the slopes in all dimensions
    if compute_jacobian:
        logj = torch.log(torch.prod(slopes, 1))

    del slopes

    # Compute the integral over the left-bins.
    # 1. Compute all integrals: cumulative sum of bin height * bin weight.
    # We want that index i contains the cumsum *strictly to the left* so we shift by 1
    # leaving the first entry null, which is achieved with a roll and assignment
    q_left_integrals = torch.roll(torch.cumsum(q, 2) * w, 1, 2)
    q_left_integrals[:, :, 0] = 0

    # 2. Access the correct index to get the left integral of each point and add it to our transformation
    out = out + torch.gather(q_left_integrals, 2, mx.unsqueeze(-1)).squeeze(-1)

    # Regularization: points must be strictly within the unit hypercube
    # Use the dtype information from pytorch
    eps = torch.finfo(out.dtype).eps
    out = out.clamp(
        min=eps,
        max=1. - eps
    )

    return out, logj


def piecewise_linear_inverse_transform(y, q_tilde, compute_jacobian=True):
    """
    Apply the inverse of an element-wise piecewise-linear transformation to some variables

    Parameters
    ----------
    y : torch.Tensor
        a tensor with shape (N,k) where N is the batch dimension while k is the
        dimension of the variable space. This variable span the k-dimensional unit
        hypercube

    q_tilde: torch.Tensor
        is a tensor with shape (N,k,b) where b is the number of bins.
        This contains the un-normalized heights of the bins of the piecewise-constant PDF for dimension k,
        i.e. q_tilde lives in all of R and we don't impose a constraint on their sum yet.
        Normalization is imposed in this function using softmax.

    compute_jacobian : bool, optional
        determines whether the jacobian should be compute or None is returned

    Returns
    -------
    tuple of torch.Tensor
        pair `(x,h)`.
        - `x` is a tensor with shape (N,k) living in the k-dimensional unit hypercube
        - `j` is the jacobian of the transformation with shape (N,) if compute_jacobian==True, else None.
    """

    # TODO do a bottom-up assesment of how we handle the differentiability of variables

    # Compute the bin width w
    N, k, b = q_tilde.shape
    Ny, ky = y.shape
    assert N == Ny and k == ky, "Shape mismatch"

    w = 1. / b

    # Compute the normalized bin heights by applying a softmax function on the bin dimension
    q = 1. / w * third_dimension_softmax(q_tilde)

    # Compute the integral over the left-bins in the forward transform.
    # 1. Compute all integrals: cumulative sum of bin height * bin weight.
    # We want that index i contains the cumsum *strictly to the left* so we shift by 1
    # leaving the first entry null, which is achieved with a roll and assignment
    q_left_integrals = torch.roll(torch.cumsum(q, 2) * w, 1, 2)
    q_left_integrals[:, :, 0] = 0

    # We can figure out which bin each y belongs to by finding the smallest bin such that
    # y - q_left_integral is positive

    edges = (y.unsqueeze(-1) - q_left_integrals).detach()
    # y and q_left_integrals are between 0 and 1 so that their difference is at most 1.
    # By setting the negative values to 2., we know that the smallest value left
    # is the smallest positive
    edges[edges < 0] = 2.
    edges = torch.clamp(torch.argmin(edges, dim=2), 0, b - 1).to(torch.long)

    # Need special error handling because trying to index with mx
    # if it contains nans will lock the GPU. (device-side assert triggered)
    if torch.any(torch.isnan(edges)).item() or torch.any(edges < 0) or torch.any(edges >= b):
        raise AvertedCUDARuntimeError("NaN detected in PWLinear bin indexing")

    # Gather the left integrals at each edge. See comment about gathering in q_left_integrals
    # for the unsqueeze
    q_left_integrals = q_left_integrals.gather(2, edges.unsqueeze(-1)).squeeze(-1)

    # Gather the slope at each edge.
    q = q.gather(2, edges.unsqueeze(-1)).squeeze(-1)

    # Build the output
    x = (y - q_left_integrals) / q + edges * w

    # Regularization: points must be strictly within the unit hypercube
    # Use the dtype information from pytorch
    eps = torch.finfo(x.dtype).eps
    x = x.clamp(
        min=eps,
        max=1. - eps
    )

    # Prepare the jacobian
    logj = None
    if compute_jacobian:
        logj = - torch.log(torch.prod(q, 1))
    return x.detach(), logj


class ElementWisePWLinearTransform(InvertibleTransform):
    """Invertible piecewise-linear transformations over the unit hypercube

    Implements a batched bijective transformation `h` from the d-dimensional unit hypercube to itself,
    in an element-wise fashion (each coordinate transformed independently)

    In each direction, the bijection is a piecewise-linear transform with b bins
    where the forward transform has evenly spaced bins. The transformation in each bin is
    actually an affine transformation. The slopes for each direction and each point in the batch
    are given by an unormalized tensor `q_tilde`. This input is softmax-normalized such that
    1. h(0) = 0
    2. h(1) = 1
    3. h is monotonous
    4. h is continuous

    for which knowing the slopes in each bin is sufficient (when the abuse of language "linear")

    Conditions 1. to 3. ensure the transformation is a bijection and therefore invertible
    The inverse is also an element-wise, piece-wise linear transformation,
    but, of course, with variable input bin sizes (and fixed output bin sizes).
    """

    backward = staticmethod(piecewise_linear_transform)
    forward = staticmethod(piecewise_linear_inverse_transform)


class GeneralPWLinearCoupling(InvertibleCouplingCell):
    """Abstract class implementing a coupling cell based on PW linear transformations

    A specific way to predict the parameters of the transform must be implemented
    in child classes.
    """

    def __init__(self, *, d, mask):
        """Generator for the abstract class GeneralPWLinearCoupling

        Parameters
        ----------
        d: int
            dimension of the space
        mask: list of bool
            variable mask which variables are transformed (False)
            or used as parameters of the transform (True)

        """
        super(GeneralPWLinearCoupling, self).__init__(d=d, transform=ElementWisePWLinearTransform(), mask=mask)


class PWLinearCoupling(GeneralPWLinearCoupling):
    """Piece-wise Linear coupling

    Coupling cell using an element-wise piece-wise linear transformation as a change of
    variables. The transverse neural network is a rectangular dense neural network

    Notes:
        Transformation used:
        `zunis.models.flows.coupling_cells.piecewise_coupling.piecewise_linear.ElementWisePWLinearTransform`
        Neural network used:
        zunis.models.layers.trainable.ArbitraryShapeRectangularDNN
    """

    def __init__(self, *, d, mask,
                 n_bins=10,
                 d_hidden=256,
                 n_hidden=8,
                 input_activation=Reshift,
                 hidden_activation=torch.nn.LeakyReLU,
                 output_activation=None,
                 use_batch_norm=False):
        """
        Generator for PWLinearCoupling

        Parameters
        ----------
        d: int
        mask: list of bool
            variable mask: which dimension are transformed (False) and which are not (True)
        n_bins: int
            number of bins in each dimensions
        d_hidden: int
            dimension of the hidden layers of the DNN
        n_hidden: int
            number of hidden layers in the DNN
        input_activation: optional
            pytorch activation function before feeding into the DNN.
            must be a callable generator without arguments (i.e. a classname or a function)
        hidden_activation: optional
            pytorch activation function between hidden layers of the DNN.
            must be a callable generator without arguments (i.e. a classname or a function)
        output_activation: optional
            pytorch activation function at the output of the DNN.
            must be a callable generator without arguments (i.e. a classname or a function)
        use_batch_norm: bool
            whether batch normalization should be used in the DNN.
        """

        super(PWLinearCoupling, self).__init__(d=d, mask=mask)

        d_in = sum(mask)
        d_out = d - d_in

        self.T = ArbitraryShapeRectangularDNN(d_in=d_in,
                                              out_shape=(d_out, n_bins),
                                              d_hidden=d_hidden,
                                              n_hidden=n_hidden,
                                              input_activation=input_activation,
                                              hidden_activation=hidden_activation,
                                              output_activation=output_activation,
                                              use_batch_norm=use_batch_norm)



import logging
import torch
from copy import deepcopy
from collections import OrderedDict
from math import isfinite, ceil
from better_abc import ABC, abstractmethod, abstract_attribute
from .training_record import TrainingRecord
from zunis.utils.logger import set_verbosity as set_verbosity_fct
from zunis.utils.exceptions import AvertedCUDARuntimeError, NoCheckpoint, TrainingInterruption


class InvalidLossError(ValueError):
    """Value error that indicates something is wrong with a loss value specifically"""
    pass


class GenericTrainerAPI(ABC):
    """Weighted dataset trainer API definition

    The goal of this API specification is to interact with the Integrator
    """

    @abstractmethod
    def train_on_batch(self, x, px, fx, **kwargs):
        """Training function

        Arguments:
            - x:
                batch of points in target space sampled from a distribution p(x)
            - px:
                the corresponding batch of p(x) values
            - fx:
                the (un-normalized) target weights
            - kwargs:
                trainer-specific options. This overrides the config

        Notes:
            This is not enforced at the level of this API but it is intended that kwargs leads to a call of
            set_config and therefore that these options are saved - at least by default.
        """

    @abstractmethod
    def set_config(self, **kwargs):
        """Set training configuration options"""

    @abstractmethod
    def get_config(self, **kwargs):
        """Get the current training configuration"""

    @abstractmethod
    def sample_forward(self, n_points):
        """Sample points using the model"""

    def reset(self):
        """Reinitinalize trainer and model. Optional element of the API.
        Raises an error if not implemented
        """
        raise NotImplementedError("reset method not available for this class")


class BasicTrainer(ABC):
    """Basic trainer implementation: sample points in target space from a fixed distribution and train over
    a fixed number of epochs on each batch, over a fixed number of batches.

    This is the implementation of all training facilities for this training mode at a low level: no automation,
    tracking, checkpointing etc is performed. No state or history is conserved.

    Rationale: the training implementation should be independent from the tracking and from the API
    """

    set_verbosity = set_verbosity_fct

    def __init__(self, flow, latent_prior, verbosity=None):
        self.flow = flow
        self.latent_prior = latent_prior
        self.loss = abstract_attribute()
        self.logger = logging.getLogger(__name__).getChild(self.__class__.__name__ + ":" + hex(id(self)))
        self.set_verbosity(verbosity)

    def sample_forward(self, n_points):
        if self.flow.inverse:
            self.flow.invert()

        xj = self.latent_prior(n_points)
        with torch.no_grad():
            xj = self.flow(xj)
        return xj.detach()

    def process_loss(self, loss):
        if not isfinite(loss):
            raise InvalidLossError(f"loss value {loss} is not valid")
        self.logger.debug(f"Loss: {loss:.3e}")
        return False

    def handle_invalid_loss(self, error):
        raise

    def handle_cuda_error(self, error):
        raise

    def compute_loss_no_grad(self, x, px, fx):
        if not self.flow.inverse:
            self.flow.invert()

        with torch.no_grad():
            xj = torch.cat(x, torch.ones(x.shape[0], 1), dim=1)

            zj = self.flow(xj)
            z = zj[:, :-1]
            logqx = zj[:, -1] + self.latent_prior.log_prob(z)
            loss = self.loss(fx, px, logqx)

        return loss.cpu().item()

    def train_step_on_target_minibatch(self, x, px, fx, optim):
        """Perform one training set on a minibatch

        Parameters
        ----------
            x:
                batch of points in target space sampled from some PDF p(x)
            px:
                values of p(x) for the batch
            fx:
                values of the target (un-normalized) PDF
            optim:
                pytorch optimizer object
        """
        if not self.flow.inverse:
            self.flow.invert()

        xj = torch.cat([x, torch.zeros(x.shape[0], 1).to(x.device)], dim=1)

        zj = self.flow(xj)
        z = zj[:, :-1]
        logqx = zj[:, -1] + self.latent_prior.log_prob(z)
        optim.zero_grad()
        loss = self.loss(fx, px, logqx)
        loss.backward()
        optim.step()
        loss = loss.detach().cpu().item()

        return loss

    def train_step_on_target_batch(self, x, px, fx, optim, minibatch_size=None):
        """Training function on a fixed batch: iterate once over the whole batch

        Parameters
        ----------
            x: torch.Tensor
                batch of points in target space sampled from some PDF p(x)
            px: torch.Tensor
                values of p(x) for the batch
            fx: torch.Tensor
                values of the target (un-normalized) PDF
            optim: torch.optim.Optimize
                pytorch optimizer object
            minibatch_size: None or int
                Optional. Size of each minibatch for gradient steps.

        Notes
        -----
            if minibatch_size is unset (or None), then it is set to the size of the full batch and a single
            gradient step is taken
        """
        n_points = x.shape[0]
        if minibatch_size is None:
            minibatch_size = n_points
        elif isinstance(minibatch_size, float) and 0 < minibatch_size <= 1.:
            minibatch_size = int(minibatch_size * x.shape[0])
        else:
            assert isinstance(minibatch_size,
                              int) and minibatch_size > 0, f"minibatch size must be None, a float in ]0,1] or a " \
                                                           f"positive int. Got {type(minibatch_size)}"

        begin = 0
        while begin < n_points:
            end = min(begin + minibatch_size, n_points)
            loss = self.train_step_on_target_minibatch(
                x[begin:end],
                px[begin:end],
                fx[begin:end],
                optim,
            )
            early_stop = self.process_loss(loss)
            if early_stop:
                raise TrainingInterruption("Early stopping condition was raised")
            begin = end

    def train_on_target_batch(self, x, px, fx, optim, n_epochs, minibatch_size=None):
        """Training function on a fixed batch: train for a fixed number of epochs on each batch

        Parameters
        ----------
            x:
                batch of points in target space sampled from some PDF p(x)
            px:
                values of p(x) for the batch
            fx:
                values of the target (un-normalized) PDF
            optim:
                pytorch optimizer object
            n_epochs:
                number of iterations over the full batch
            minibatch_size:
                Optional. Size of each minibatch for gradient steps

        Notes
        -----
            if minibatch_size is unset (or None), then it is set to the size of the full batch and a single
            gradient step is taken
        """
        self.logger.info(f"Training on batch: {x.shape[0]} points")
        for epoch in range(n_epochs):
            self.logger.info(f"Epoch {epoch + 1}/{n_epochs}")
            try:
                self.train_step_on_target_batch(x, px, fx, optim, minibatch_size)
            except InvalidLossError as e:
                self.logger.exception("Invalid loss detected - handling started")
                self.handle_invalid_loss(e)
            except AvertedCUDARuntimeError as e:
                self.logger.exception("CUDA error averted - handling started")
                self.handle_cuda_error(e)

    @staticmethod
    def generate_target_batch_from_posterior(n_points, f, target_posterior):
        """Generate a batch of training examples in target space from a specified distribution

        Parameters
        ----------
            n_points:
                size of the batch
            f:
                function to evaluate on the sampled points
            target_posterior:
                distribution from which to sample points in target_space

        Returns
        -------
            tuple of torch.Tensor
                (x,px,fx): sampled points, sampling distribution PDF values, function values
        """
        xlpx = target_posterior(n_points)
        x = xlpx[:, :-1]
        px = torch.exp(- xlpx[:, -1])
        fx = f(x)

        return x, px, fx

    def train_on_target_batches_from_posterior(self,
                                               *,
                                               f,
                                               batch_size,
                                               n_batches=1,
                                               n_epochs_per_batch,
                                               minibatch_size=None,
                                               target_posterior,
                                               optim):
        """Main training function: iterate over a fixed number of batches sampled in target space and
         train for a fixed number of epochs on each batch

        Parameters
        ----------
            f:
                un-normalized target PDF
            batch_size:
                number of points per batch
            n_batches:
                number of batches to train over
            n_epochs_per_batch:
                number of iterations over each full batch before sampling a new one
            minibatch_size:
                Optional. Size of each minibatch for gradient steps
            target_posterior: distribution from which points are sampled in target space
            optim:
                pytorch optimizer object

        Notes
        -----
            if minibatch_size is unset (or None), then it is set to the size of the full batch and a single
            gradient step is taken
        """
        self.logger.info(f"Training on {n_batches} independent batches of size {batch_size}")
        self.logger.info(f"Performing {n_epochs_per_batch} epochs per batch")
        self.logger.info(f"Each epoch applies {int(ceil(batch_size / minibatch_size))} gradient steps")

        for batch in range(n_batches):
            self.logger.info("Generating a new batch of points")
            x, px, fx = self.generate_target_batch_from_posterior(batch_size, f, target_posterior)
            self.train_on_target_batch(x, px, fx, optim, n_epochs_per_batch, minibatch_size)


class BasicStatefulTrainer(BasicTrainer, GenericTrainerAPI):
    def __init__(self, flow, latent_prior, checkpoint=True, checkpoint_on_cuda=True, checkpoint_path=None, max_reloads=None,
                 **kwargs):
        super(BasicStatefulTrainer, self).__init__(flow, latent_prior)

        # Setting up the saved configuration accessed through the GenericTrainerAPI
        # Note that the GenericTrainerAPI, operates at the level of a single batch
        # As a result, only relevant keys are part of the config

        # The keys are stored separately so as to be immutable
        self.config_keys = ("n_epochs", "minibatch_size", "optim")

        self.config = {
            "n_epochs": None,
            "minibatch_size": None,
            "optim": None
        }

        for key in kwargs:
            if key in self.config_keys:
                self.config[key] = kwargs[key]

        # Checkpointing logic
        self.checkpoint = checkpoint
        self.checkpoint_on_cuda = checkpoint_on_cuda
        self.checkpoint_data = None
        self.checkpoint_path = checkpoint_path
        self.n_reloads = 0
        if max_reloads is None:
            self.max_reloads = None if not checkpoint else 1
        else:
            self.max_reloads = max_reloads

        if self.checkpoint_path is None:
            self.record = TrainingRecord()
        else:
            self.record = TrainingRecord(checkpoint=checkpoint_path)

    def set_checkpoint(self):
        """Save the current model state as a checkpoint"""
        if not self.checkpoint:
            self.logger.error("This trainer cannot save checkpoints")
            raise AssertionError("This trainer cannot save checkpoints")

        state_dict = deepcopy(self.flow.state_dict())
        if self.checkpoint_on_cuda:
            self.checkpoint_data = state_dict
        else:
            self.checkpoint_data = OrderedDict(
                [(key, value.cpu()) for key, value in state_dict.items()]
            )
        if self.checkpoint_path is not None:
            torch.save(self.checkpoint_data, self.checkpoint_path)

    def restore_checkpoint(self, path=None):
        """Restore from a checkpoint if available"""
        # This function is called when something fails
        # as a result, it should always raise its specific type of error: NoCheckpoint so that the
        # more important failure cause can be reported
        # We use logging.exception to still report the local error stack trace in logs

        # If a path is provided, use it and fail otherwise
        # Use a NoCheckpoint exception to allow whatever error triggered this function to be the main failure cause
        if path is not None:
            try:
                self.flow.load_state_dict(torch.load(path))
                return
            except Exception:
                self.logger.exception(f"Error when loading state dict from file")
                raise NoCheckpoint(f"Could not load checkpoint from {path}")

        # Default functioning mode: no path specified, use initialization settings
        # We have to have set the options to keep checkpoints in memory and have saved at least once
        if self.checkpoint and self.checkpoint_data is not None:
            self.logger.warning("Reloading the latest checkpoint from memory")
            try:
                self.flow.load_state_dict(self.checkpoint_data)
                return
            # In case of error, log the stack trace and try loading from disk if possible
            except Exception:
                self.logger.exception("Could not load checkpoint from memory")
            if self.checkpoint_path is not None:
                self.logger.warning("Trying to load latest checkpoint from disk")
                try:
                    self.flow.load_state_dict(torch.load(self.checkpoint_path))
                    return
                # Again, if we fail, log the stack trace and raise a NoCheckpoint to allow upstream triage
                except Exception:
                    self.logger.exception("Error when loading state dict from file")
                    raise NoCheckpoint(f"Could not load checkpoint from {self.checkpoint_path}")

        raise NoCheckpoint("No checkpoint available")

    def restore_checkpoint_except(self):
        """Try to to restore from a checkpoint as a response to an exception"""
        if "checkpoint" in self.record and self.n_reloads < self.max_reloads:
            self.logger.warning(f"Attempting checkpoint reload {self.n_reloads + 1}/{self.max_reloads}")
            self.restore_checkpoint()
            self.n_reloads += 1
            return True
        else:
            self.logger.error("Cannot reset to a previous checkpoint to sidestep exception")
            return False

    def handle_invalid_loss(self, error):
        """In case of invalid loss, we try to reload the latest checkpoint"""
        if self.restore_checkpoint_except():
            return

        # Otherwise reload the checkpoint and re-raise the error
        try:
            self.logger.warning("An InvalidLossError could not be sidestepped. Trying to reload checkpoint before interrupting the training")
            self.restore_checkpoint()
            self.logger.warning("Could load the latest checkpoint")
            interruptor = TrainingInterruption("Invalid loss error interrupted training")
            self.logger.error(repr(interruptor))
            raise interruptor
        except NoCheckpoint:
            pass

        super(BasicStatefulTrainer, self).handle_invalid_loss(error)

    def handle_cuda_error(self, error):
        """In case of a cuda error, we try to reload the latest checkpoint"""
        if self.restore_checkpoint_except():
            return

        # Otherwise reload the checkpoint and re-raise the error
        try:
            self.logger.warning("An AvertedCUDARuntimeError could not be sidestepped. Trying to reload checkpoint before interrupting the training")
            self.restore_checkpoint()
            self.logger.warning("Could load the latest checkpoint")
            interruptor = TrainingInterruption("CUDA error interrupted training")
            self.logger.error(repr(interruptor))
            raise interruptor
        except NoCheckpoint:
            pass

        super(BasicStatefulTrainer, self).handle_cuda_error(error)

    def process_loss(self, loss):
        """Handle invalid losses by reloading checkpoint and handle logging and checkpointing if valid"""

        # parent class checks that the loss is valid
        output = super(BasicStatefulTrainer, self).process_loss(loss)

        # Handle logging and checkpointing
        self.record.log_loss(loss)
        if self.checkpoint and self.record["loss"] <= self.record["best_loss"]:
            self.set_checkpoint()
        return output

    def train_step_on_target_minibatch(self, x, px, fx, optim):
        loss = super(BasicStatefulTrainer, self).train_step_on_target_minibatch(x, px, fx, optim)
        self.record.next_step()
        return loss

    def train_step_on_target_batch(self, x, px, fx, optim, minibatch_size=None):
        self.record.new_epoch()
        super(BasicStatefulTrainer, self).train_step_on_target_batch(x, px, fx, optim, minibatch_size)

    def train_on_target_batches_from_posterior(self,
                                               *,
                                               f,
                                               batch_size,
                                               n_batches=1,
                                               n_epochs_per_batch,
                                               minibatch_size=None,
                                               target_posterior,
                                               optim):
        """Train over several epochs over several batches with explicit arguments - overriding the current config
        We save the full config of this run here and therefore have a different config format
        """

        config = {
            "f": f,
            "batch_size": batch_size,
            "n_batches": n_batches,
            "n_epochs_per_batch": n_epochs_per_batch,
            "minibatch_size": minibatch_size,
            "target_posterior": target_posterior,
            "optim": optim
        }

        self.record = TrainingRecord(config=config)
        super(BasicStatefulTrainer, self).train_on_target_batches_from_posterior(**config)
        return self.record

    def set_config(self, **kwargs):
        """Set the saved configuration
        """
        for key in kwargs:
            if key in self.config_keys:
                self.config[key] = kwargs[key]
            else:
                self.logger.error(f"Specified config key {key} is not recognized")
                raise KeyError(f"Config key {key} was not recognized")

    def get_config(self):
        """Return the saved configuration"""
        return self.config

    def train_on_batch(self, x, px, fx, **kwargs):
        """Train on a batch of points using the saved configuration"""
        self.set_config(**kwargs)

        try:
            checkpoint_path = kwargs["checkpoint_path"]
        except KeyError:
            checkpoint_path = self.checkpoint_path

        self.record = TrainingRecord(checkpoint=checkpoint_path)

        optim = self.config["optim"]
        if optim is None:
            error = ValueError("trainer parameter 'optim' must be set before or at training")
            self.logger.error(repr(error))
            raise error

        n_epochs = self.config["n_epochs"]
        self.logger.debug(f"n_epochs in trainer: {n_epochs}")
        if n_epochs is None:
            error = ValueError("trainer parameter 'n_epochs' must be set before or at training")
            self.logger.error(repr(error))
            raise error

        minibatch_size = self.config["minibatch_size"]
        self.train_on_target_batch(x, px, fx, optim=optim, n_epochs=n_epochs, minibatch_size=minibatch_size)
        return self.record



"""This training module provides functions to train invertible flow models on datasets
drawn from a known distribution, which does not correspond to the target of our flow model,
which we know as a black box function

In other words, we have a set of (x,p(x),q(x)), such that
- x lives in the target space
- x ~ p(x)
- q is a black box function
- we want to train a flow model F such that F(z) ~ q(F(z)) when we sample z in the latent space

Note that typical cases where a flow is trained on an experimental dataset
- q = p
- p is unknown
In which case one can maximize the likelihood of the dataset under the flow model.
"""


"""ZuNIS: Neural importance sampling library"""
import logging
import sys
from zunis.integration import integration_logger as ilogger
from zunis.training import training_logger as tlogger


logger = logging.getLogger(__name__)
"""Overall parent logger for the Zunis library"""
logger_integration = ilogger
"""Overall parent logger for all integration operations"""
logger_training = tlogger
"""Overall parent logger for all training operations"""

# Ensure that no submodule loggers outputs anything, unless explicitly setup by the user
logger.addHandler(logging.NullHandler())


def setup_std_stream_logger(min_level=None, debug=False, force=True):
    """Set the root logger of this module to split output between stdout and stderr
    Arguments:
        - min_level: minimum level output to stderr. Default is INFO
        - debug: overrides min_level to DEBUG

    Notes:
        stdout : [min_level, WARNING)
        stderr : [WARNING, +∞)
    """

    if min_level is None:
        min_level = logging.INFO

    if debug:
        min_level = logging.DEBUG

    class InfoFilter(logging.Filter):
        def filter(self, rec):
            return min_level <= rec.levelno < logging.WARNING

    logger.setLevel(min_level)

    h1 = logging.StreamHandler(sys.stdout)
    h1.setLevel(min_level)
    h1.addFilter(InfoFilter())
    h2 = logging.StreamHandler(sys.stderr)
    h2.setLevel(logging.WARNING)

    if len(logger.handlers) == 1 or force:
        logger.addHandler(h1)
        logger.addHandler(h2)
    else:
        logger.warning("The logger already has handlers setup none were added.")
        logger.warning("Use the `force` option to bypass this behavior.")



import logging
from better_abc import ABC, abstractmethod, abstract_attribute
import torch

from zunis.utils.logger import set_verbosity as set_verbosity_fct
from zunis.utils.exceptions import TrainingInterruption

class SurveyRefineIntegratorAPI(ABC):
    """API specification for an integrator that performs a number of survey steps in which
    a flow model is trained and then a number of refine steps in which sampling is done through the model
    The interaction with the model is done through the GenericTrainerAPI
    """

    set_verbosity = set_verbosity_fct

    def __init__(self, *args, verbosity=None, **kwargs):
        self.model_trainer = abstract_attribute()
        self.logger = logging.getLogger(__name__).getChild(self.__class__.__name__ + ":" + hex(id(self)))
        self.set_verbosity(verbosity)

    @abstractmethod
    def initialize(self, **kwargs):
        """Intialization before the whole integration process"""

    @abstractmethod
    def initialize_refine(self, **kwargs):
        """Initialization before the survey phase"""

    @abstractmethod
    def initialize_survey(self, **kwargs):
        """Initialization before the survey phase"""

    @abstractmethod
    def sample_survey(self, **kwargs):
        """Sample points for a survey step"""

    @abstractmethod
    def sample_refine(self, **kwargs):
        """Sample points for a refine step"""

    @abstractmethod
    def process_survey_step(self, sample, integral, integral_var, **kwargs):
        """Process the result of a survey step"""

    @abstractmethod
    def process_refine_step(self, sample, integral, integral_var, **kwargs):
        """Process the result of a refine step"""

    @abstractmethod
    def finalize_survey(self, **kwargs):
        """Perform the final operations of the survey phase"""

    @abstractmethod
    def finalize_refine(self, **kwargs):
        """Perform the final operations of the refine phase"""

    @abstractmethod
    def finalize_integration(self, **kwargs):
        """Perform the final operations of the whole integration"""

    def format_arguments(self, **kwargs):
        """Format keyword arguments passed to the train function in a suitable way
        In this generic API definition, the kwargs must be passed in the correct format
        however for a specific realization, options should be made accessible through keywords
        which will then be sorted into the right structure.
        """
        return kwargs

    def survey_step(self, **kwargs):
        """Basic survey step: sample points, estimate the integral, its error, train model

        possible keyword arguments:
            sampling_args: dict
            training_args: dict
        """

        try:
            sampling_args = kwargs["sampling_args"]
        except KeyError:
            sampling_args = dict()

        try:
            training_args = kwargs["training_args"]
        except KeyError:
            training_args = dict()

        x, px, fx = self.sample_survey(**sampling_args)
        integral_var, integral = torch.var_mean(fx / px)
        integral = integral.cpu().item()
        integral_var = integral_var.cpu().item()

        training_record = self.model_trainer.train_on_batch(x, px, fx, **training_args)

        self.process_survey_step((x, px, fx), integral, integral_var, training_record=training_record)

    def refine_step(self, **kwargs):
        """Basic refine step: sample points, estimate the integral, its error, train model

        possible keyword arguments:
            sampling_args: dict
        """

        try:
            sampling_args = kwargs["sampling_args"]
        except KeyError:
            sampling_args = dict()

        x, px, fx = self.sample_refine(**sampling_args)
        integral_var, integral = torch.var_mean(fx / px)
        integral = integral.cpu().item()
        integral_var = integral_var.cpu().item()

        self.process_refine_step((x, px, fx), integral, integral_var)

    def survey(self, n_survey_steps=10, **kwargs):
        """Perform a survey phase of integration

        possible keyword arguments:
            trainer_config_args: dict
            survey_step_args: dict
            finalize_survey_args: dict

        """
        self.logger.info("Initializing the survey phase")

        try:
            trainer_config_args = kwargs["trainer_config_args"]
            self.model_trainer.set_config(**trainer_config_args)
        except KeyError:
            pass

        try:
            survey_step_args = kwargs["survey_step_args"]
        except KeyError:
            survey_step_args = dict()

        try:
            initialize_survey_args = kwargs["initialize_survey_args"]
        except KeyError:
            initialize_survey_args = dict()

        try:
            finalize_survey_args = kwargs["finalize_survey_args"]
        except KeyError:
            finalize_survey_args = dict()

        self.initialize_survey(**initialize_survey_args)

        self.logger.info("Starting the survey phase")
        try:
            for i in range(n_survey_steps):
                self.survey_step(**survey_step_args)
        except TrainingInterruption as e:
            self.logger.debug(" "*72)
            self.logger.debug("="*72)
            self.logger.debug(" "*20 + "/!\ This message is important" + " "*23)
            self.logger.debug("="*72)
            self.logger.exception("Survey interrupted by a training interruption")
            self.logger.error("A working checkpoint was loaded successfully and integration can go on.")

        self.logger.info("Finalizing the survey phase")
        self.finalize_survey(**finalize_survey_args)

    def refine(self, n_refine_steps=10, **kwargs):
        """Perform the refine phase of integration

        Possible keyword arguments:

        trainer_config_args: dict
        refine_step_args: dict
        finalize_refine_args: dict
        """

        self.logger.info("Initializing the refine phase")

        try:
            trainer_config_args = kwargs["trainer_config_args"]
            self.model_trainer.set_config(**trainer_config_args)
        except KeyError:
            pass

        try:
            refine_step_args = kwargs["refine_step_args"]
        except KeyError:
            refine_step_args = dict()

        try:
            initialize_refine_args = kwargs["initialize_refine_args"]
        except KeyError:
            initialize_refine_args = dict()

        try:
            finalize_refine_args = kwargs["finalize_refine_args"]
        except KeyError:
            finalize_refine_args = dict()

        self.initialize_refine(**initialize_refine_args)

        self.logger.info("Starting the refine phase")
        for i in range(n_refine_steps):
            self.refine_step(**refine_step_args)

        self.logger.info("Finalizing the refine phase")
        self.finalize_refine(**finalize_refine_args)

    def integrate(self, n_survey_steps=10, n_refine_steps=10, **kwargs):
        self.logger.info("Starting integration")

        formated_kwargs = self.format_arguments(**kwargs)

        try:
            survey_args = formated_kwargs["survey_args"]
        except KeyError:
            survey_args = dict()

        try:
            refine_args = formated_kwargs["refine_args"]
        except KeyError:
            refine_args = dict()

        try:
            finalize_args = formated_kwargs["finalize_args"]
        except KeyError:
            finalize_args = dict()

        self.survey(n_survey_steps, **survey_args)
        self.refine(n_refine_steps, **refine_args)

        return self.finalize_integration(**finalize_args)



from math import sqrt
from better_abc import abstractmethod
import torch
from .flat_survey_integrator import FlatSurveySamplingIntegrator


class AdaptiveSurveyIntegrator(FlatSurveySamplingIntegrator):
    """Adaptive integrator based on a separation between survey and sampling

    Survey:
        Sample points and spend some time training the model

        Sampling is done in two phases:

        1. Sample in the target space (input space of the integrand) using a uniform distribution
        2. Sample in the latent space and use the model to sample point in the target space


        The switch between the two phases is performed based on a test method - abstract here - that checks
        whether the flat distribution does a better job of estimating the loss than the flat distribution

    Refine:
        Sample points using the trained model and evaluate the integral
    """

    def __init__(self, f, trainer, d, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 device=torch.device("cpu"), verbosity=2, trainer_verbosity=1, **kwargs):
        super(AdaptiveSurveyIntegrator, self).__init__(f,
                                                       trainer,
                                                       d,
                                                       n_iter=n_iter,
                                                       n_iter_survey=n_iter_survey,
                                                       n_iter_refine=n_iter_refine,
                                                       n_points=n_points,
                                                       n_points_survey=n_points_survey,
                                                       n_points_refine=n_points_refine,
                                                       use_survey=use_survey,
                                                       device=device,
                                                       verbosity=verbosity,
                                                       trainer_verbosity=trainer_verbosity,
                                                       **kwargs)

        # As long as the survey switch condition (see abstractmethod below) is False,
        # This is false
        self.sample_forward = False

    @abstractmethod
    def survey_switch_condition(self):
        """Boolean valued method that checks if it is time to switch between sampling uniformly
        and using the model"""

    def sample_survey(self, *, n_points=None, f=None, **kwargs):

        # Starting mode: sample from the flat distribution in target space
        if not self.sample_forward:
            return super(AdaptiveSurveyIntegrator, self).sample_survey(n_points=n_points, f=f, **kwargs)

        # When the model is trained enough, sample from it
        if n_points is None:
            n_points = self.n_points_survey
        if f is None:
            f = self.f

        xj = self.model_trainer.sample_forward(n_points)
        x = xj[:, :-1]
        px = torch.exp(- xj[:, -1])
        fx = f(x)

        return x, px, fx

    def process_survey_step(self, sample, integral, integral_var, training_record, **kwargs):
        super(AdaptiveSurveyIntegrator, self).process_survey_step(sample, integral, integral_var, training_record,
                                                                  **kwargs)
        if (not self.sample_forward) and self.survey_switch_condition():
            self.logger.info("Switching sampling mode")
            self.sample_forward = True


class ForwardSurveySamplingIntegrator(AdaptiveSurveyIntegrator):
    """Survey/Refine integrator using forward sampling for training during the survey phase.

    The constructor option `ForwardSurveySamplingIntegrator` (default: `False`) allows using flat latent space sampling
    for the first survey step, then switching to forward sampling.
    """

    def __init__(self, f, trainer, d, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 device=torch.device("cpu"), verbosity=2, trainer_verbosity=1, sample_flat_once=False, **kwargs):
        super(ForwardSurveySamplingIntegrator, self).__init__(f,
                                                              trainer,
                                                              d,
                                                              n_iter=n_iter,
                                                              n_iter_survey=n_iter_survey,
                                                              n_iter_refine=n_iter_refine,
                                                              n_points=n_points,
                                                              n_points_survey=n_points_survey,
                                                              n_points_refine=n_points_refine,
                                                              use_survey=use_survey,
                                                              device=device,
                                                              verbosity=verbosity,
                                                              trainer_verbosity=trainer_verbosity,
                                                              **kwargs)

        self.sample_forward = not sample_flat_once

    def survey_switch_condition(self):
        return True


class VarianceAdaptiveSurveyIntegrator(AdaptiveSurveyIntegrator):
    """Survey/Refine adaptive integrator based on the variance loss.
    The integrator estimates the variance of a flat integrator and switches to forward sampling if
    the flow performs significantly (2 sigma) better.
    """

    def __init__(self, f, trainer, d, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 device=torch.device("cpu"), verbosity=2, trainer_verbosity=1, **kwargs):
        self.flat_loss = None
        self.flat_loss_std = None
        super(VarianceAdaptiveSurveyIntegrator, self).__init__(f,
                                                               trainer,
                                                               d,
                                                               n_iter=n_iter,
                                                               n_iter_survey=n_iter_survey,
                                                               n_iter_refine=n_iter_refine,
                                                               n_points=n_points,
                                                               n_points_survey=n_points_survey,
                                                               n_points_refine=n_points_refine,
                                                               use_survey=use_survey,
                                                               device=device,
                                                               verbosity=verbosity,
                                                               trainer_verbosity=trainer_verbosity,
                                                               **kwargs)

    def compute_flat_variance_loss(self, fx):
        """Compute the variance loss and its standard deviation assuming points are sampled from a flat distribution
        We clip the loss standard deviation to loss/4 so that the switching condition can be met
        """
        losses = fx ** 2
        self.flat_loss = losses.mean().cpu().item()
        self.flat_loss_std = min(self.flat_loss/4., losses.std().cpu().item())

    def survey_switch_condition(self):
        return self.model_trainer.record["loss"] < self.flat_loss - self.flat_loss_std

    def process_survey_step(self, sample, integral, integral_var, training_record, **kwargs):
        if not self.sample_forward:
            _, _, fx = sample
            self.compute_flat_variance_loss(fx)

        super(VarianceAdaptiveSurveyIntegrator, self).process_survey_step(sample, integral, integral_var,
                                                                          training_record,
                                                                          **kwargs)



import logging


verbosity_levels = {0: logging.ERROR,
                    1: logging.WARNING,
                    2: logging.INFO,
                    3: logging.DEBUG,
                    "ERROR": logging.ERROR,
                    "WARNING": logging.WARNING,
                    "INFO": logging.INFO,
                    "DEBUG": logging.DEBUG,
                    }


def set_verbosity(obj, level):
    if level is None:
        return
    level_key = level
    if isinstance(level, str):
        level_key = level.upper()
    try:
        obj.logger.setLevel(verbosity_levels[level_key])
    except KeyError:
        obj.logger.warning(f"Verbosity level {level} not recognized. Ignored.")



"""Integration package using neural importance sampling

The main API is exposed as :py:func:`zunis.integration.Integrator <zunis.integration.default_integrator.Integrator>`
"""
import logging
from zunis.integration.default_integrator import Integrator

integration_logger = logging.getLogger(__name__)
"""Overall parent logger for all integration operations"""



"""Tools to build out-of-the-box trainers"""
import torch
from .dkl_training import BasicStatefulDKLTrainer
from zunis.models.generators import create_hypercube_flow
from zunis.models.flows.sampling import FactorizedGaussianSampler


def create_dkl_trainer(d, model_type="realnvp", mask_type="checkerboard", model_params=None,
                       device=torch.device("cpu"), n_epochs=10, optim_class=None, lr=1.e-4, **params):
    """Create a dkl trainer with the a default flow model"""
    if model_params is None:
        model_params = dict()

    model = create_hypercube_flow(d, model_type=model_type, mask_type=mask_type, **model_params).to(device)
    prior = FactorizedGaussianSampler(d=d, device=device)

    if optim_class is None:
        optim = torch.optim.Adam(model.parameters(), lr=lr)
    else:
        optim = optim_class(model.parameters(),lr=lr)

    return BasicStatefulDKLTrainer(model, prior, optim=optim, n_epochs=n_epochs, **params)



from setuptools import find_packages, setup

setup(
    name='zunis',
    packages=find_packages(where='zunis_lib'),
    package_dir={'': 'zunis_lib'},
    install_requires=[
        "numpy == 1.19.1",
        "pandas == 1.1.0",
        "torch == 1.6.0",
        "abc_property == 1.0",
        "dictwrapper == 1.3",
        "ruamel.yaml <= 0.16.12"
    ],
    version='0.3rc1',
    description='Neural Importance Sampling',
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author='Nicolas Deutschmann',
    author_email="nicolas.deutschmann@gmail.com",
    url="https://zunis.readthedocs.io",
    download_url="https://github.com/ndeutschmann/zunis",
    license='MIT',
    include_package_data = True
)



"""Activation functions
Element-wise transformations without trainable parameters
"""
import torch

class BiLU(torch.nn.Module):
    """Bijective linear unit
    f(x) = alpha + delta*sign(x)
    """

    def __init__(self, alpha=1., delta=.9):
        super(BiLU, self).__init__()
        self.alpha = torch.nn.Parameter(torch.tensor(alpha), requires_grad=False)
        self.delta = torch.nn.Parameter(torch.tensor(delta), requires_grad=False)

    def forward(self, input):
        """Output of the BiLU activation function"""
        return input * self.alpha + self.delta * torch.abs(input)


class NormBiTanh(torch.nn.Module):
    """Bijective Normalized Tanh layer
     f(x) = alpha*Tanh + (1-alpha)*x activation
     """

    def __init__(self, alpha=0.3):
        super(NormBiTanh, self).__init__()
        self.alpha = torch.nn.Parameter(torch.tensor(alpha), requires_grad=False)

    def forward(self, input):
        """Output of the NormBiTanh activation function"""
        return self.alpha * input + (1 - self.alpha) * torch.tanh(input)



"""Implementation of realNVP: a coupling cell with an affine transform"""

import torch
from .general_coupling import InvertibleCouplingCell
from zunis.models.layers.trainable import ArbitraryShapeRectangularDNN
from .transforms import InvertibleTransform


def element_wise_affine(x, st, compute_jacobian=True):
    """Transform x element-wise through an affine function y = exp(s)*x + t
    where s = st[...,0] and t = st[...,1] with s.shape == x.shape == t.shape

    The Jacobian for this transformation is the coordinate-wise product of the scaling factors
    J = prod(es[...,i],i)
    """
    es = torch.exp(st[..., 0])
    t = st[..., 1]
    logj = None
    if compute_jacobian:
        logj = torch.sum(torch.log(es), dim=-1)

    return es * x + t, logj


def inverse_element_wise_affine(x, st, compute_jacobian=True):
    """Transform x element-wise through an affine function y = exp(-s)*(x - t)
    where s = st[...,0] and t = st[...,1] with s.shape == x.shape == t.shape
    This is the inverse of `element_wise_affine` above for the same set of parameters st

    The Jacobian for this transformation is the coordinate-wise product of the scaling factors
    J = prod(es[...,i],i)
    """
    es = torch.exp(-st[..., 0])
    t = st[..., 1]
    logj = None
    if compute_jacobian:
        logj = torch.sum(torch.log(es), dim=-1)

    return es * (x - t), logj


class ElementWiseAffineTransform(InvertibleTransform):
    """Invertible element-wise affine transform"""
    def forward(self, y, st, compute_jacobian=True):
        """"""
        return element_wise_affine(y, st, compute_jacobian=compute_jacobian)

    def backward(self, y, st, compute_jacobian=True):
        """"""
        return inverse_element_wise_affine(y, st, compute_jacobian=compute_jacobian)


class GeneralRealNVP(InvertibleCouplingCell):
    """Coupling cell based on affine transforms without a specific parameter prediction function"""

    def __init__(self, *, d, mask):
        super(GeneralRealNVP, self).__init__(d=d, transform=ElementWiseAffineTransform(), mask=mask)


class FakeT(torch.nn.Module):
    """Fake neural net for the test FakeRealNVP that returns a trainable constant"""

    def __init__(self, s=0., t=0.):
        super(FakeT, self).__init__()
        self.st = torch.nn.Parameter(torch.tensor([s, t]), requires_grad=True)

    def forward(self, x):
        return self.st


class FakeRealNVP(GeneralRealNVP):
    """Fake real NVP with position independent affine parameters
    for basic debugging and training tests
    """

    def __init__(self, *, d, mask, s=0., t=0.):
        super(FakeRealNVP, self).__init__(d=d, mask=mask)
        self.T = FakeT(s, t)


class RealNVP(GeneralRealNVP):
    """Real non-volume-preserving coupling cell.

    This implements a bijective mapping R^d -> R^d by transforming some subset of the coordinates
    defined by the mask through a coordinate-wise affine transformation. As with other coupling cells
    the parameters of the affine transformation are the output of a neural network taking the un-tranformed
    coordinates as input."""
    def __init__(self, *, d, mask,
                 d_hidden=256,
                 n_hidden=8,
                 input_activation=None,
                 hidden_activation=torch.nn.LeakyReLU,
                 output_activation=None,
                 use_batch_norm=False):
        """

        Parameters
        ----------
        d: int
            dimension of the space
        mask: list of bool
            mask defining which variables are transformed
        d_hidden: int, 256
            width of the neural network
        n_hidden: int, 8
            depth of the neural network
        input_activation
            activation function applied before going through the neural network
        hidden_activation
            activation function applied after each hidden layer of the neural network
        output_activation
            activation function applied after the last layer of the neural network
        use_batch_norm: bool
            whether to use batch normalization after each hidden layer of the neural network
        """
        super(RealNVP, self).__init__(d=d, mask=mask)

        d_in = sum(mask)
        d_out = d - d_in

        self.T = ArbitraryShapeRectangularDNN(d_in=d_in,
                                              out_shape=(d_out, 2),
                                              d_hidden=d_hidden,
                                              n_hidden=n_hidden,
                                              input_activation=input_activation,
                                              hidden_activation=hidden_activation,
                                              output_activation=output_activation,
                                              use_batch_norm=use_batch_norm)



"""Sampling layers
The input of flows as we use them is nearly always generated data from some distribution
provided with its log-inverse PDF. As a result, it can be convenient to plug the first layer
of a flow as a sampling layers that draws points and computes the required PDF
"""

import torch


class FactorizedFlowSampler(torch.nn.Module):
    """Sample d-dimensional data from a factorized 1D PDF over each dimension
    The 1D PDF is expected to be a pytorch.probability.Distribution object
    but it can be any object that implements the `sample`and `log_prob` methods.

    NB: we provide the 1D prior object explicitly as pytorch distributions don't
    respond appropriately to the `.to(device)`. To sample on a device, provide a
    prior initialized with parameters already on the correct device.
    """
    def __init__(self, *, d, prior_1d):
        super(FactorizedFlowSampler, self).__init__()
        self.d = d
        self.prior = prior_1d

    def log_prob(self, x):
        """Compute, point-per-point, the log-PDF of a batch of points"""
        assert len(x.shape) == 2 and x.shape[1] == self.d, f"Expected shape: (:, {self.d})"
        return torch.sum(self.prior.log_prob(x), -1)

    def forward(self, n_batch):
        """Sample n_batch points and stack them with their jacobians"""
        # Generate normally distributed points
        x = self.prior.sample((n_batch, self.d))
        # Compute their per-dimension log PDF and sum them across dimenions
        # to get the log PDF then add a batch dimension
        log_j = - self.log_prob(x)
        return torch.cat([x, log_j.unsqueeze(-1)], -1)


class FactorizedGaussianSampler(FactorizedFlowSampler):
    """Factorized gaussian prior
    Note that tensorflow distribution objects cannot easily be moved devices so specify the right
    device at initialization.
    """
    def __init__(self, *, d, mu=0., sig=1., device=None):
        # Copy data
        sig_ = torch.tensor(sig)
        mu_ = torch.tensor(mu)

        if device is not None:
            sig_ = sig_.to(device)
            mu_ = mu_.to(device)

        prior_1d = torch.distributions.normal.Normal(mu_, sig_)
        super(FactorizedGaussianSampler, self).__init__(d=d, prior_1d=prior_1d)


class UniformSampler(FactorizedFlowSampler):
    """Factorized uniform prior
    Note that tensorflow distribution objects cannot easily be moved devices so specify the right
    device at initialization.
    """
    def __init__(self, *, d, low=0., high=1., device=None):
        # Copy data
        low_ = torch.tensor(low, dtype=torch.get_default_dtype())
        high_ = torch.tensor(high, dtype=torch.get_default_dtype())

        if device is not None:
            low_ = low_.to(device)
            high_ = high_.to(device)

        prior_1d = torch.distributions.Uniform(low_, high_)
        super(UniformSampler, self).__init__(d=d, prior_1d=prior_1d)



import torch

# TODO: change where the method we use from this class is
from zunis.integration.base_integrator import BaseIntegrator
from zunis.training.weighted_dataset.weighted_dataset_trainer import BasicStatefulTrainer
from zunis.models.flows.sampling import UniformSampler


class PosteriorSurveySamplingIntegrator(BaseIntegrator):
    """Integrator using a target space posterior to sample during survey:
    Takes a trainer object at initialization and samples from a given distribution in target space
    during the survey phase. The function to integrate is given at initialization.
    """

    def __init__(self, f, trainer, posterior, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 verbosity=None, trainer_verbosity=None, **kwargs):
        super(PosteriorSurveySamplingIntegrator, self).__init__(f=f,
                                                                trainer=trainer,
                                                                trainer_verbosity=trainer_verbosity,
                                                                n_iter=n_iter,
                                                                n_iter_survey=n_iter_survey,
                                                                n_iter_refine=n_iter_refine,
                                                                n_points=n_points,
                                                                n_points_survey=n_points_survey,
                                                                n_points_refine=n_points_refine,
                                                                use_survey=use_survey,
                                                                verbosity=verbosity,
                                                                **kwargs)

        self.posterior = posterior

    def sample_survey(self, *, n_points=None, f=None, **kwargs):
        """Sample points from target space distribution"""
        # TODO, change where this method is /!\
        if n_points is None:
            n_points = self.n_points_survey
        if f is None:
            f = self.f
        return BasicStatefulTrainer.generate_target_batch_from_posterior(n_points, f, self.posterior)


class FlatSurveySamplingIntegrator(PosteriorSurveySamplingIntegrator):
    def __init__(self, f, trainer, d, n_iter=10, n_iter_survey=None, n_iter_refine=None,
                 n_points=100000, n_points_survey=None, n_points_refine=None, use_survey=False,
                 device=torch.device("cpu"), verbosity=2, trainer_verbosity=1, **kwargs):
        posterior = UniformSampler(d=d, device=device)
        super(FlatSurveySamplingIntegrator, self).__init__(f=f,
                                                           trainer=trainer,
                                                           d=d,
                                                           posterior=posterior,
                                                           n_iter=n_iter,
                                                           n_iter_survey=n_iter_survey,
                                                           n_iter_refine=n_iter_refine,
                                                           n_points=n_points,
                                                           n_points_survey=n_points_survey,
                                                           n_points_refine=n_points_refine,
                                                           use_survey=use_survey,
                                                           verbosity=verbosity,
                                                           trainer_verbosity=trainer_verbosity,
                                                           **kwargs)
        self.d = d



"""Implementation of the abstract GeneralFlow class
Most generic variable transformation:
- takes in a point x and -log(PDF(x))
- outputs a transformed point y and - log(PDF(y)) = - log(PDF(x)) + log(dy/dx)

Reminder:

dx p(x) = dy q(y) = dx dy/dx q(y)
=> q(y) = p(x)/(dy/dx)
=> -log q(y) = -log p(x) + log dy/dx
"""
import torch
from better_abc import ABC, abstract_attribute, abstractmethod


class GeneralFlow(torch.nn.Module,ABC):
    """General abstract class for flows"""

    def __init__(self, *, d):
        super(GeneralFlow, self).__init__()
        self.d = d

    @abstractmethod
    def flow(self, x):
        """Transform the batch of points x  with shape (...,d)
        This is an abstract method that should be overriden"""
        pass

    @abstractmethod
    def transform_and_compute_jacobian(self, xj):
        """Compute the flow transformation and its Jacobian simulatenously on
        xj with xj.shape == (...,d+1)

        This is an abstract method that should be overriden
        """
        pass

    def forward(self, xj):
        """Compute the flow transformation on some input xj
        - In training mode, xj.shape == (:,d+1)
        and the last dimension is the log-inverse PDF of x[:,:-1]
        - In eval mode,, xj.shape == (:,d)
        and no jacobian is passed: pure sampling mode.
        """
        if self.training:
            assert xj.shape[-1] == self.d+1
            return self.transform_and_compute_jacobian(xj)
        else:
            assert xj.shape[-1] == self.d
            return self.flow(xj)







import os
from functools import partial

import torch
from zunis.utils.config import Configuration

from .defaults import local_path


def get_optim_from_config(config=None):
    """Get an optimizer object from a config dictionary

    Parameters
    ----------
    config: dictionary-like, optional
        dictionary-like object with the following structure:
        {
        "optim_cls": <optim class name within torch.optim>,
        "optim_config": <dictionary defining the key-word argument parameters for instantiating that class>
        }
        If no argument is given, the default configuration is used (utils/config/optim_config.yaml)

    Returns
    -------
        :obj:`function`
            a pre-filled optimizer constructor that only needs to be fed the model parameters.

    Notes
    -----
    This function does not instantiate the optimize object, it wraps its constructor with functools.partial
    to pre-fill the optimizer settings with the provided value. Given a model, one then needs to call the output
    on model.parameters()
    """
    if config is None:
        config = Configuration.from_yaml(os.path.join(local_path, "optim_config.yaml"))

    optim = getattr(torch.optim, config["optim_cls"])
    return partial(optim, **config["optim_config"].as_dict())


def get_default_integrator_config():
    """Get a copy of the full default text-based integrator config

    The default configuration file is situated in utils/config/integrator_config.yaml
    """
    return Configuration.from_yaml(os.path.join(local_path, "integrator_config.yaml"))


def create_integrator_args(config=None):
    """Create the full hierarchy of arguments for an integrator, including an instantiated optimizer

    Parameters
    ----------
    config: dictionary-like, optional
        Full configuration dictionary for the Integrator function. If none is provided, the default configuration
        is used.

    Returns
    -------
        :obj:dict
            keyword dictionary ready to be provided as ``**kwargs`` to the
            :py:func:`Integrator <zunis.integration.default_integrator.Integrator>` function
    """
    if config is None:
        config = get_default_integrator_config()

    try:
        optim_config = config["optim"]
    except KeyError:
        optim_config = get_optim_from_config()
    args = config.as_dict()
    try:
        args["trainer_options"]["optim"] = get_optim_from_config(optim_config)
    except KeyError:
        args["trainer_options"] = {"optim": get_optim_from_config(optim_config)}
    return args


"""Flows based on coupling cells

A coupling cell is a variable change x^i = f(y^j) such that there are two variable groups
M and N such that for all n in N and m in M
* x^n = y^n
* x^m = f(x^m,T(x^N)) <- each m transformed indepently and depends on m and all N
such that f(x,t^N) has a simple derivative wrt x.
"""


"""Function wrappers to provide API-compatible functions"""

import torch


def wrap_numpy_hypercube_batch_function(f):
    """Take a function that evaluates on numpy batches with values in the unit hypercube
    and return a function that evaluates on pytorch batches in the unit hypercube
    """
    def torchf(x):
        npx = x.detach().cpu().numpy()
        npfx = f(npx)
        return torch.tensor(npfx, device=x.device)

    return torchf


def wrap_numpy_compact_batch_function(f, dimensions):
    """Take a function that evaluates on numpy batches with values in compact intervals provided
    as a list of shape (d,2) where each element is the pair (lower,upper) of interval boundaries.
    and return a function that evaluates on pytorch batches in the unit hypercube,
    weighted by the proper Jacobian factor to preserve integrals.
    """
    tdim = torch.tensor(dimensions)
    assert tdim.shape[1] == 2, "argument dimensions is expected to have shape (N,2)"
    assert torch.all(tdim[:, 1] > tdim[:, 0]), "Each dimension is expected to be (a,b) with a<b"
    starts = tdim[:, 0]
    lengths = tdim[:, 1] - tdim[:, 0]
    jac = torch.prod(lengths).cpu().item()

    def torchf(x):
        npx = (x * lengths.to(x.device) + starts.to(x.device)).detach().cpu().numpy()
        npfx = f(npx)
        return (torch.tensor(npfx) * jac).to(x.device)

    return torchf


def wrap_compact_arguments_function(f, dimensions):
    """Take a function that evaluates on a sequence of arguments with values in compact intervals provided
    as a list of shape (d,2) where each element is the pair (lower,upper) of interval boundaries and
    return a function that evaluates on pytorch batches in the unit hypercube,
    weighted by the proper Jacobian factor to preserve integrals.

    Explicitly: f(x_1,x_2,...,x_N) where x_i are numbers in [dimensions[i][0], dimensions[i][1]] returns a single float.
    """
    tdim = torch.tensor(dimensions)
    assert tdim.shape[1] == 2, "argument dimensions is expected to have shape (N,2)"
    assert torch.all(tdim[:, 1] > tdim[:, 0]), "Each dimension is expected to be (a,b) with a<b"
    starts = tdim[:, 0]
    lengths = tdim[:, 1] - tdim[:, 0]
    jac = torch.prod(lengths).item()

    def torchf(x):
        lxs = (x * lengths.to(x.device) + starts.to(x.device)).detach().cpu().tolist()
        fxs = torch.zeros(x.shape[0], device=x.device)
        for i, lx in enumerate(lxs):
            fxs[i] = f(*lx)
        return fxs * jac

    return torchf


def wrap_hypercube_arguments_function(f):
    """Take a function that evaluates on a sequence of arguments with values in the unit hypercube and
        return a function that evaluates on pytorch batches in the unit hypercube,
        weighted by the proper Jacobian factor to preserve integrals.

        Explicitly: f(x_1,x_2,...,x_N) where x_i are numbers in [0, 1] returns a single float.
        """
    def torchf(x):
        lxs = x.detach().cpu().tolist()
        fxs = torch.zeros(x.shape[0], device=x.device)
        for i, lx in enumerate(lxs):
            fxs[i] = f(*lx)
        return fxs

    return torchf



"""Abstract class for coupling cells"""
import torch
from better_abc import abstract_attribute

from zunis.models.flows.general_flow import GeneralFlow
from .transforms import InvertibleTransform


def not_list(l):
    """Return the element wise negation of a list of booleans"""
    assert all([isinstance(it, bool) for it in l])
    return [not it for it in l]


class GeneralCouplingCell(GeneralFlow):
    """Abstract class for coupling cell"""

    def __init__(self, *, d, transform, mask):
        """
        Parameters
        ----------
        d: int
            dimension of the space
        transform: function or :py:class:`InvertibleTransform <zunis.models.flows.coupling_cells.transforms.InvertibleTransform>`
            bijective variable transform that maps the unit hypercube (in `d-sum(mask)` dimensions) to itself
            given some parameters. It should have the same signature as
            :py:meth:`InvertibleTransform.forward <zunis.models.flows.coupling_cells.transforms.InvertibleTransform.forward>`
        mask: list of bool
            indicates which variables are passed (y_N) through and which are transformed (y_M)
        """
        super(GeneralCouplingCell, self).__init__(d=d)

        self.mask = mask + [False]
        self.mask_complement = not_list(mask) + [False]
        self.transform = transform
        self.T = abstract_attribute()

    def transform_and_compute_jacobian(self, yj):
        """Apply the variable change on a batch of points y and compute the jacobian"""
        y_n = yj[..., self.mask]
        y_m = yj[..., self.mask_complement]
        log_j = yj[..., -1]

        x = torch.zeros_like(yj).to(yj.device)
        x[..., self.mask] = y_n
        x[..., self.mask_complement], log_jy = self.transform(y_m, self.T(y_n), compute_jacobian=True)
        x[..., -1] = log_j + log_jy
        return x

    def flow(self, y):
        """Apply the variable change on a batch of points y"""
        y_n = y[..., self.mask[:-1]]
        y_m = y[..., self.mask_complement[:-1]]

        x = torch.zeros_like(y).to(y.device)
        x[..., self.mask] = y_n
        x[..., self.mask_complement], _ = self.transform(y_m, self.T(y_n), compute_jacobian=False)

        return x


class InvertibleCouplingCell(GeneralCouplingCell):
    """An invertible coupling cell works exactly the same as a general coupling cell
    but its transform admits and inverse transformation.
    """

    def __init__(self, *, d, transform, mask):
        """
        Parameters
        ----------
        d: int
            dimension of the space
        transform: :py:class:`InvertibleTransform <zunis.models.flows.coupling_cells.transforms.InvertibleTransform>`
            bijective variable transform that maps the unit hypercube (in `d-sum(mask)` dimensions) to itself
            given some parameters.
        mask: list of bool
            indicates which variables are passed (y_N) through and which are transformed (y_M)"""
        assert isinstance(transform, InvertibleTransform)
        super(InvertibleCouplingCell, self).__init__(d=d, transform=transform, mask=mask)

    def invert(self):
        """Change the running mode from forward to inverse (invert the change of variable)"""
        self.transform.invert()

    def runs_forward(self):
        """Check the running mode: True if forward and False if backward/inverse"""
        return not self.transform.inverse
